# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections import ChainMap
from collections.abc import Mapping
from functools import cached_property

from pydantic import BaseModel, Field, computed_field


class FileConfig(BaseModel):
    format_: str = Field(alias='format')
    header: bool
    name: str
    submitter: str

    comment: str = '#'
    default_reference_header: str | None = Field(
        None,
        alias='default reference header',
    )


class DestinationConfig(BaseModel):
    key: str

    imported_values: list[str | dict[str, str]] = []
    separator: str = '|'
    ignore: list[str] = []

    @computed_field(return_type=list[type])
    @cached_property
    def magic_types(self) -> list[type]:
        type_map = {
            'int': int,
            'double': float,
        }

        return [
            v
            for k, v in type_map.items()
            if k in self.imported_values
        ]

    @computed_field(return_type=bool)
    @property
    def magic_match_all(self) -> bool:
        return 'all' in self.imported_values

    @computed_field(return_type=dict[str, str])
    @cached_property
    def imported_values_map(self) -> dict[str, str]:
        maps = [
            a
            for a
            in self.imported_values
            if isinstance(a, Mapping)
        ]
        chained_map = dict(
            ChainMap(*maps)
        )
        literal_dict = {
            a: a
            for a in self.imported_values
            if not isinstance(a, Mapping)
        }

        return chained_map | literal_dict


class AttributeConfig(BaseModel):
    imported_column_name: str
    column_reference: str
    destination: DestinationConfig | list[DestinationConfig]


class YamlConfig(BaseModel):
    file: FileConfig
    attributes: list[AttributeConfig] = Field(
        min_length=1,
    )



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
from os import path

from marshmallow import Schema, fields

import requests

from ..core import DataSource


class Interface(DataSource):
    def __init__(self, config):
        # pmb_url baracoda_url, generate_limit, print_limit
        super().__init__(config)

    def custom_response(self, status_code=200, message='SUCCESS', data={}):
        return {'status_code': status_code, 'message': message, 'data': data}

    def printers(self):
        """Show all printers"""
        printers = []
        try:
            response = requests.get(path.join(self.pmb_url, 'v1', 'printers'))
        except requests.exceptions.RequestException as error:
            return self.custom_response(status_code=421, message=str(error))
        for i in response.json()['data']:
            printers += [
                {
                    'id': i['id'],
                    'name': i['attributes']['name'],
                    'type': i['attributes']['printer_type'],
                }
            ]
        return self.custom_response(data=printers)

    def label_templates(self):
        """Show all label-templates"""
        try:
            response = requests.get(path.join(self.pmb_url, 'v1', 'label_templates'))
        except requests.exceptions.RequestException as error:
            return self.custom_response(status_code=421, message=str(error))
        if response.status_code != 200:
            raise self.custom_response(
                status_code=response.status_code, data=response.json()
            )
        label_templates = []
        for i in response.json()['data']:
            label_templates += [{'id': i['id'], 'name': i['attributes']['name']}]
        return self.custom_response(data=label_templates)

    def required_fields(self, label_template_name):
        """Show all required fields for a label template"""
        label_templates = {
            label['name']: label['id'] for label in self.label_templates()['data']
        }
        _id = label_templates[label_template_name]
        try:
            response = requests.get(
                path.join(self.pmb_url, 'v1', 'label_templates', _id)
            )
        except requests.exceptions.RequestException as error:
            return self.custom_response(status_code=421, message=str(error))
        required_fields = ['label_name']
        for i in response.json()['included']:
            if i['type'] == 'bitmaps':
                required_fields += [i['attributes']['field_name']]
        return self.custom_response(data=required_fields)

    def generate(self, prefix, number):
        """Generate barcodes with given prefix"""
        if number > self.generate_limit:
            return self.custom_response(
                403,
                message=f'Requested to generate more barcodes than limit of {self.generate_limit}'
            )
        request = path.join(
            self.baracoda_url, 'barcodes_group', str(prefix), f'new?count={number}'
        )
        try:
            response = requests.post(request)
        except requests.exceptions.RequestException as error:
            return self.custom_response(status_code=421, message=str(error))
        if response.status_code != 201:
            return self.custom_response(
                status_code=response.status_code,
                message='Baracoda error',
                data=response,
            )
        try:
            barcodes = response.json()['barcodes_group']['barcodes']
            return self.custom_response(data=barcodes)
        except requests.exceptions.RequestException as error:
            return self.custom_response(status_code=421, message=str(error))

    def _label_schema(self, label_template_name):
        """Marshmallow schema from list of required fields"""
        required_fields_response = self.required_fields(label_template_name)
        required_fields = required_fields_response['data']
        schema_dict = {'barcode': fields.Str(required=True)}
        for field in required_fields:
            schema_dict[field] = fields.Str(required=True)
        schema_dict.pop('label_name')
        return Schema.from_dict(schema_dict)

    def print_labels(
        self, label_data, printer_name, label_template_name, copies=1, dry=True
    ):
        """Print labels"""
        if copies > self.print_limit:
            return self.custom_response(
                403,
                message=f'Requested to print more barcodes than limit of {self.print_limit}'
            )

        copies = min(copies, self.print_limit)
        schema = self._label_schema(label_template_name)

        for label in label_data:
            validation = schema().validate(label)
            if validation != {}:
                return self.custom_response(
                    status_code=400, message='Validation error', data=validation
                )

        for label in label_data:
            label['label_name'] = 'main_label'

        job = {
            'print_job': {
                'printer_name': printer_name,
                'label_template_name': label_template_name,
                'labels': label_data,
                'copies': copies,
            }
        }

        url = path.join(self.pmb_url, 'v2', 'print_jobs')

        if not dry:
            try:
                response = requests.post(
                    url,
                    data=json.dumps(job),
                    headers={
                        'Content-Type': 'application/vnd.api+json',
                        'accept': 'application/json',
                    },
                )
                return self.custom_response(
                    status_code=response.status_code,
                    message=response.text
                )
            except requests.exceptions.RequestException as error:
                return self.custom_response(status_code=421, message=str(error))

        return self.custom_response(message='Dry run')

    @property
    def supported_types(self):
        raise NotImplementedError()

    @property
    def attribute_types(self):
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class StsSpeciesToElasticSpeciesConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        lab_work_statuses = []
        statuses = {}
        for species_lab_work_status in data_object.species_lab_work_statuses:
            lab_work_statuses.append(
                species_lab_work_status.status
            )
            statuses[species_lab_work_status.status.lower() + '_date'] = \
                species_lab_work_status.updated_at

        if len(lab_work_statuses) > 0:
            statuses['lab_work_status'] = lab_work_statuses

        if data_object.sequencing_material_status is not None:
            statuses['sequencing_material_status'] = data_object.sequencing_material_status.status

        ret = self._data_object_factory(
            'species',
            data_object.id,
            attributes=data_object.attributes | statuses,
        )
        yield ret



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import os
import typing
from dataclasses import dataclass
from datetime import datetime
from typing import Any, List

from flask import Blueprint, request

from .auth import ForbiddenError
from .misc import (
    CtxGetter,
    default_ctx_getter
)
from ..core import (
    DataSourceError,
    DataSourceFilter
)

if typing.TYPE_CHECKING:
    from ..prefect import PrefectDataSource
    from ..sql import SqlDataSource


@dataclass
class UploadData:
    s3_bucket: str
    s3_filename: str
    spreadsheet_config: str
    pipeline_id: int
    destination: str
    upload_name: str
    user_id: int
    dry_run: bool


REQUIRED_FIELDS: List = [
    's3_bucket',
    's3_filename',
    'spreadsheet_config',
    'pipeline_id',
    'upload_name',
    'destination',
    'dry_run'
]


def pipeline_steps_blueprint(
    sql_ds: SqlDataSource,
    prefect_ds: PrefectDataSource,
    role: str | None = None,
    url_prefix: str = '/run-pipeline',

    ctx_getter: CtxGetter = default_ctx_getter,
) -> Blueprint:
    """
    A flask `Blueprint` providing endpoints
    for running validations on uploaded manifests.

    E.g. running a Tree of Sex validation pipeline
    on a spreadsheet manifest using `PrefectDataSource`.
    """

    bp = Blueprint(
        'pipeline-steps',
        __name__,
        url_prefix=url_prefix
    )

    def __check_required_fields(
        body: dict[str, Any],
        required_fields: list[str] = REQUIRED_FIELDS
    ) -> None:

        if not all(field in body for field in required_fields):
            raise DataSourceError(
                'Bad Request',
                f'You must specify all of: {", ".join(required_fields)}',
                400
            )

    def __get_pipeline(
        pipeline_id: int
    ) -> str:

        pipeline = sql_ds.get_one(
            'pipeline',
            pipeline_id
        )

        if not pipeline:
            raise DataSourceError(
                'Not Found',
                'The specified pipeline was not found.',
                404
            )

        return pipeline.id

    def __insert_upload_data(
        upload_data: UploadData
    ) -> str:

        upload = sql_ds.data_object_factory(
            'upload',
            attributes={
                's3_bucket': upload_data.s3_bucket,
                's3_filename': upload_data.s3_filename,
                'spreadsheet_config': upload_data.spreadsheet_config,
                'upload_name': upload_data.upload_name,
                'pipeline_id': upload_data.pipeline_id,
                'destination': upload_data.destination,
                'user_id': upload_data.user_id,
                'validation_status': 'validation_in_progress'
            },
        )

        inserted_upload_data = list(
            sql_ds.insert(
                'upload',
                [upload]
            )
        )

        if not inserted_upload_data:
            raise DataSourceError(
                'Insertion Error',
                'Failed to insert upload data.',
                500
            )

        return inserted_upload_data[0].id

    def __insert_flow_run(
        upload_id: str,
        pipeline_id: str,
        s3_filename: str,
        dry_run: bool = False,
        flow_name: str = 'pipeline',
    ) -> str | None:

        flow_params = {
            'upload_id': upload_id,
            'pipeline_id': pipeline_id,
            'dry_run': dry_run,
            'source_kwargs': {
                's3_filename': s3_filename,
                's3_bucket': os.environ['UPLOAD_S3_BUCKET'],
            }
        }

        flow_run = prefect_ds.data_object_factory(
            'flow_run',
            attributes={
                'flow_name': flow_name,
                'deployment_name': flow_name,
                'parameters': flow_params,
                'tags': [
                    f'app_name: {os.environ.get("APP_NAME", "tol")}',
                ],
            }
        )

        inserted_flow_data = list(
            prefect_ds.insert(
                'flow_run',
                [flow_run]
            )
        )[0]
        return inserted_flow_data.id

    def __upsert_flow_run_id(
        upload_id: str,
        flow_run_id: str
    ) -> None:

        upload = sql_ds.data_object_factory(
            'upload',
            upload_id,
            attributes={
                'flow_run_id': flow_run_id
            }
        )

        sql_ds.upsert(
            'upload',
            [upload]
        )

    def __check_auth(ctx: Any = None) -> None:
        if role is not None:
            if not ctx or not ctx.authenticated:
                raise ForbiddenError()
            if role not in ctx.roles:
                raise ForbiddenError()

    @bp.route('/revalidate', methods=['POST'])
    def revalidate_upload() -> tuple[dict[str, str], int]:

        allowed_validation_statuses = ['validation_system_error', 'validation_timeout']

        ctx = ctx_getter()
        __check_auth(ctx)

        body: dict[str, Any] = request.json.get('data', {})
        __check_required_fields(body, required_fields=['upload_ids'])

        upload_ids: list[str] = body['upload_ids']

        f = DataSourceFilter()
        f.and_ = {
            'id': {'in_list': {'value': upload_ids}}}

        existing_uploads = sql_ds.get_list(
            'upload',
            object_filters=f
        )

        if not existing_uploads:
            raise DataSourceError(
                'Not Found',
                'No uploads found for the provided upload IDs.',
                404
            )

        existing_by_id = {upload.id: upload for upload in existing_uploads}

        for upload in existing_uploads:
            if upload.validation_status not in allowed_validation_statuses:
                raise DataSourceError(
                    'Invalid Request',
                    f'Upload with ID {upload.id} cannot be revalidated as its validation \
                    status is not in {allowed_validation_statuses} status.',
                    400
                )

        uploads = [sql_ds.data_object_factory(
            'upload',
            upload_id,
            attributes={
                'validation_status': 'validation_in_progress',
                'rejection_reason': None,
                'date_started': datetime.now(),
                'failure_message': None,
                'validation_results': [],
                'completed': False,
                'flow_run_id': None
            },
        ) for upload_id in upload_ids]

        if not uploads:
            raise DataSourceError(
                'Not Found',
                'No uploads found for the provided upload IDs.',
                404
            )

        sql_ds.upsert(
            'upload',
            uploads
        )

        flow_run_ids = [__insert_flow_run(
            upload_id=upload_id,
            pipeline_id=existing_by_id[upload_id].pipeline_id,
            s3_filename=existing_by_id[upload_id].s3_filename,
            dry_run=False
        ) for upload_id in upload_ids]

        for upload_id, flow_run_id in zip(upload_ids, flow_run_ids):
            __upsert_flow_run_id(
                upload_id=upload_id,
                flow_run_id=flow_run_id
            )

        return {'success': True, 'upload_and_flow_run_ids': list(zip(upload_ids,
                                                                     flow_run_ids))}, 200

    @bp.post('')
    def run_pipeline_steps() -> tuple[dict[str, Any], int]:

        ctx = ctx_getter()
        __check_auth(ctx)

        user_id = ctx.user_id if ctx and ctx.authenticated else None

        body: dict[str, Any] = request.json.get('data', {})

        __check_required_fields(body)

        pipeline_id: str = body['pipeline_id']

        __get_pipeline(pipeline_id)

        upload_data = UploadData(
            s3_bucket=body['s3_bucket'],
            s3_filename=body['s3_filename'],
            spreadsheet_config=body['spreadsheet_config'],
            pipeline_id=pipeline_id,
            destination=body['destination'],
            upload_name=body['upload_name'],
            user_id=user_id,
            dry_run=body['dry_run']
        )

        upload_id = body.get('upload_id') if body.get('upload_id') is not None \
            else __insert_upload_data(upload_data)

        flow_run_id = __insert_flow_run(
            upload_id=upload_id,
            pipeline_id=pipeline_id,
            s3_filename=upload_data.s3_filename,
            dry_run=upload_data.dry_run
        )

        __upsert_flow_run_id(
            upload_id=upload_id,
            flow_run_id=flow_run_id
        )

        return {'success': True, 'upload_id': upload_id, 'flow_run_id': flow_run_id}, 200

    return bp



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT
from abc import ABC, abstractmethod
from typing import Any


class AuthorisationManager(ABC):
    """
    Abstract base class for managing user authorisations and memberships.

    This class defines the interface for authorisation managers that need to
    retrieve user membership information for access control purposes. Concrete
    implementations should handle the specific logic for querying user memberships
    from various authorisation systems (e.g., LDAP, database, external services).

    The class follows the ABC (Abstract Base Class) pattern to ensure that all
    concrete implementations provide the required methods for authorisation
    management.
    """

    @abstractmethod
    def get_user_memberships(self, user: Any) -> list[str]:
        """
        Retrieve the list of group memberships for a given user.

        This method should be implemented by concrete subclasses to fetch
        the user's membership information from the appropriate authorisation
        source. The memberships are typically used to determine what resources
        or actions the user is authorised to access.

        Args:
            user (Any): The user object or identifier for which to retrieve
                       memberships. The type is flexible to accommodate
                       different user representation formats (e.g., user ID,
                       user object, username string, etc.).

        Returns:
            list[str]: A list of group names or membership identifiers that
                      the user belongs to. Returns an empty list if the user
                      has no memberships or if the user is not found.

        Raises:
            NotImplementedError: If called on the abstract base class directly
                               rather than on a concrete implementation.

        Note:
            Concrete implementations may raise additional exceptions specific
            to their authorisation backend (e.g., connection errors, authentication
            failures, etc.).
        """



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import importlib.resources
import itertools
from typing import Dict, Iterable, List

import psycopg2
import psycopg2.extras
from psycopg2.extensions import connection

from ..core import (
    DataObject,
    DataSource,
    DataSourceConfig,
    DataSourceError,
    DataSourceFilter
)
from ..core.operator import ListGetter


class BenchlingWarehouseDataSource(DataSource, ListGetter):
    """
    A (read-only) DataSource for getting objects in Benchling
    The queries are maintained in this SDK as SQL files
    """

    def __init__(self, config: DataSourceConfig) -> None:
        super().__init__(
            config,
            [
                'username',
                'password',
                'database',
                'hostname',
                'port',
                'schema'
            ]
        )
        self.connection = self._get_connection()

    def __get_primary_keys(self):
        return {
            'sample': 'sts_id',
            'sequencing_request': 'sanger_sample_id',
            'extraction': 'extraction_id',
            'extraction_container': 'fluidx_container_id',
            'tissue_prep': 'eln_tissue_prep_id'
        }

    def _get_connection(self) -> connection:
        return psycopg2.connect(
            user=self.username,
            password=self.password,
            database=self.database,
            port=self.port,
            host=self.hostname,
            options=f'-c search_path={self.schema},public'
        )

    def __run_query(self, sql: str) -> Iterable[DataObject]:
        with self.connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cur:
            cur.execute(sql)
            results = cur.fetchall()
            return results

    def __convert_results_to_data_objects(self, objs: Dict,
                                          object_type: str, id_col: str) -> Iterable:
        for obj in objs:
            yield self.data_object_factory(
                object_type,
                id_=obj[id_col],
                attributes={k: v for k, v in obj.items() if k != id_col}
            )

    def get_list(
        self,
        object_type: str,
        object_filters: DataSourceFilter = None,
        **kwargs
    ) -> Iterable[DataObject]:
        file_suffix = ''
        if object_filters is not None:
            if isinstance(object_filters.and_, dict):
                for field_name in [
                    'sequencing_platform', 'extraction_type', 'extraction.extraction_type'
                ]:
                    if field_name in object_filters.and_:
                        # For an in_list, treat as multiple eq
                        if 'in_list' in object_filters.and_[field_name]:
                            return itertools.chain.from_iterable(
                                self.get_list(
                                    object_type,
                                    DataSourceFilter(
                                        and_={
                                            field_name: {
                                                'eq': {
                                                    'value': val
                                                }
                                            }
                                        }
                                    )
                                ) for val in object_filters.and_[field_name]['in_list']['value']
                            )
                        elif 'eq' in object_filters.and_[field_name]:
                            file_suffix = '_' + field_name.replace('.', '_') + '_' \
                                          + object_filters.and_[field_name]['eq']['value']
            else:
                raise DataSourceError('Filtering only on sequencing platform and extraction '
                                      'type currently supported on BenchlingDataSource')
        try:
            sql = importlib.resources.files('tol.benchling.sql') \
                                     .joinpath(f'{object_type}{file_suffix}.sql') \
                                     .read_text()
            results = self.__run_query(sql)
            return self.__convert_results_to_data_objects(
                results,
                object_type,
                self.__get_primary_keys()[object_type])
        except FileNotFoundError:
            raise DataSourceError(f'Query file not found for object type: {object_type} '
                                  'with given filter')

    @property
    def supported_types(self) -> List[str]:
        return self.__get_primary_keys().keys()

    @property
    def attribute_types(self):
        return {
            t: {} for t in self.supported_types
        }



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import datetime
import logging
import sys
import uuid
from collections.abc import Mapping
from functools import cache
from typing import Dict, Iterable, Iterator

from lab_share_lib.constants import RABBITMQ_HEADER_VALUE_ENCODER_TYPE_BINARY
from lab_share_lib.rabbit.avro_encoder import AvroEncoderBinary
from lab_share_lib.rabbit.basic_publisher import BasicPublisher
from lab_share_lib.rabbit.schema_registry import SchemaRegistry
from lab_share_lib.types import RabbitServerDetails

from tol.sciops.message_builder import MessageBuilder
from tol.sciops.messages import CreateLabwareMessage, LabwareMessage, Sample

from ..core import (
    DataObject,
    DataSource,
    DataSourceError
)

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
# Change the next line to also output pika mq logging
logging.getLogger('pika').propagate = False
LOGGER = logging.getLogger(__name__)


class SequencingDataSource(DataSource):
    @property
    def registry(self):
        return self.__registry

    @property
    def rabbitmq_details(self):
        return self.__rabbitmq_details

    def __init__(self, config: Dict):
        super().__init__(config, expected=['redpanda_url', 'redpanda_api_key', 'rabbitmq_host',
                                           'rabbitmq_port', 'rabbitmq_username',
                                           'rabbitmq_password', 'rabbitmq_vhost',
                                           'rabbitmq_exchange', 'rabbitmq_routing_key',
                                           'rabbitmq_use_ssl', 'rabbitmq_publish_retry_delay',
                                           'rabbitmq_publish_retries', 'tol_feedback_queue'])
        self._initialise_rabbitmq()

    def _initialise_rabbitmq(self):
        LOGGER.info('Initialising SciOps publisher')
        self.__registry = SchemaRegistry(self.redpanda_url, self.redpanda_api_key)
        self.__rabbitmq_details = RabbitServerDetails(
            uses_ssl=self.rabbitmq_use_ssl,
            host=self.rabbitmq_host,
            port=self.rabbitmq_port,
            username=self.rabbitmq_username,
            password=self.rabbitmq_password,
            vhost=self.rabbitmq_vhost,
        )

    def _create_publisher(self) -> BasicPublisher:
        """ Create a new publisher instance for sending messages """
        return BasicPublisher(
            self.__rabbitmq_details,
            self.rabbitmq_publish_retry_delay,
            self.rabbitmq_publish_retries
        )

    def _create_encoder(self, subject) -> AvroEncoderBinary:
        """ Create a message encoder """
        encoder = AvroEncoderBinary(self.__registry, subject)
        encoder.set_compression_codec('snappy')
        return encoder

    def _send_message(self, msg_to_send: LabwareMessage):
        """ Send the given message """
        subject = msg_to_send.SUBJECT
        version = msg_to_send.VERSION
        built_msg = MessageBuilder.build_labware_message(msg_to_send)

        encoder = self._create_encoder(subject)
        encoded_message = encoder.encode([built_msg], version=version)

        try:
            publisher = self._create_publisher()
            LOGGER.info(f'Sending message {built_msg}')

            publisher.publish_message(
                self.rabbitmq_exchange,
                self.rabbitmq_routing_key,
                encoded_message.body,
                subject,
                encoded_message.version,
                RABBITMQ_HEADER_VALUE_ENCODER_TYPE_BINARY,
            )
        except Exception:
            raise DataSourceError('Unable to publish message')

    def split_objects_into_plates(self, objects):
        """ Separates the samples based on their barcode """
        separated_data = {}
        for obj in objects:
            barcode = obj.barcode
            if barcode not in separated_data:
                separated_data[barcode] = []
            separated_data[barcode].append(obj)
        return separated_data

    def _send_plate_to_sciops(
            self,
            samples: Iterable[DataObject],
            barcode: str,
            labware_type: str,
            labware_uuid: str
    ) -> None:
        """ Creates and sends the message in LabwareMessage format """
        create_msg = CreateLabwareMessage(
            message_uuid=str(uuid.uuid4()),
            message_create_date_utc=datetime.datetime.now(datetime.timezone.utc),
            barcode=barcode,
            labware_type=labware_type,
            labware_uuid=labware_uuid,
            samples=[
                Sample(
                    **{k: v for k, v in sample.attributes.items()
                       if k != 'barcode'}
                )
                for sample in samples
            ]
        )

        publisher = self
        publisher._send_message(create_msg)

    def add_plate(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        barcode: str,
        labware_type: str,
        labware_uuid: str
    ) -> None:
        """ Checks whether all samples belong to the same plate and sends message """
        first_barcode = objects[0].barcode
        for obj in objects:
            if obj.barcode != first_barcode:
                raise ValueError('Samples should have the same barcode')

        self._send_plate_to_sciops(samples=objects, barcode=barcode,
                                   labware_type=labware_type, labware_uuid=labware_uuid)
        return True

    @property
    def supported_types(self):
        return ['sequencing_sample']

    @property
    @cache
    def attribute_types(self) -> Dict:

        class AttributeTypesDict(Mapping[str, dict[str, str]]):
            def __getitem__(self, __k: str) -> dict[str, str]:
                if __k != 'sequencing_sample':
                    raise DataSourceError(
                        f'{__k} objects are not supported'
                    )
                return {
                    'barcode': 'str',
                    'sample_uuid': 'str',
                    'study_uuid': 'str',
                    'sanger_sample_id': 'str',
                    'location': 'str',
                    'supplier_sample_name': 'str',
                    'volume': 'str',
                    'concentration': 'str',
                    'public_name': 'str',
                    'taxon_id': 'str',
                    'common_name': 'str',
                    'donor_id': 'str',
                    'library_type': 'str',
                    'country_of_origin': 'str',
                    'sample_collection_date_utc': 'datetime',
                    'cost_code': 'str',
                    'genome_size': 'str',
                    'accession_number': 'str',
                    'sheared_femto_fragment_size': 'str',
                    'post_spri_concentration': 'str',
                    'post_spri_volume': 'str',
                    'final_nano_drop_280': 'str',
                    'final_nano_drop_230': 'str',
                    'final_nano_drop': 'str',
                    'shearing_and_qc_comments': 'str',
                    'date_submitted_utc': 'datetime',
                    'priority_level': 'str',
                    'date_required_by': 'str',
                    'reason_for_priority': 'str'
                }

            def __len__(self) -> int:
                return 1

            def __iter__(self) -> Iterator[str]:
                return iter(['sequencing_sample'])

        return AttributeTypesDict()



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .auth import *  # noqa
from .eln_datasource import *  # noqa
from .entities import *  # noqa
from .generators import *  # noqa
from .sanitise import *  # noqa



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any, Callable, Optional, Union

from prefect.client.schemas.filters import (
    DeploymentFilter,
    DeploymentFilterName,
    FlowFilter,
    FlowFilterName,
    FlowRunFilter,
    FlowRunFilterId,
    FlowRunFilterIdempotencyKey,
    FlowRunFilterName,
    FlowRunFilterState,
    FlowRunFilterTags
)

from ..core import DataSourceError, DataSourceFilter


CLASS_MAP = {
    'id': FlowRunFilterId,
    'idempotency_key': FlowRunFilterIdempotencyKey,
    'name': FlowRunFilterName,
    'state': FlowRunFilterState,
    'tags': FlowRunFilterTags,
    'flow_name': FlowFilterName,
    'deployment_name': DeploymentFilterName
}


_PrefectFilter = Union[
    DeploymentFilter,
    FlowFilter,
    FlowRunFilter,
]


_FilterKwargsDict = dict[str, Optional[_PrefectFilter]]


class PrefectFilterBuilder:
    """
    Provides a collection of "setter" methods, to add
    filter terms. Returns a `dict` of kwargs, using which
    to instantiate `FlowRunFilter`.
    """

    def __init__(
        self,
        flow_run_class: type[FlowRunFilter],
        deployment_class: type[DeploymentFilter],
        flow_class: type[FlowFilter],
        class_map: dict[str, type]
    ) -> None:

        self.__filters: _FilterKwargsDict = {}

        self.__flow_run_class = flow_run_class
        self.__deployment_class = deployment_class
        self.__flow_class = flow_class

        self.__class_map = class_map

    @property
    def kwargs(self) -> _FilterKwargsDict:
        return {
            'flow_run_filter': self.__flow_run_filter,
            'deployment_filter': self.__deployment_filter,
            'flow_filter': self.__flow_filter
        }

    def id_(self, any_: list[str]) -> None:
        self.__set_any('id', any_)

    def flow_name(self, any_: list[str]) -> None:
        self.__set_any('_flow_name', any_)

    def deployment_name(self, any_: list[str]) -> None:
        self.__set_any('_deployment_name', any_)

    def idempotency_key(self, any_: list[str]) -> None:
        self.__set_any('idempotency_key', any_)

    def name(self, any_: list[str]) -> None:
        self.__set_any('name', any_)

    def state(self, any_: list[str]) -> None:
        self.__set_any('state', any_)

    def tags_all(self, all_: list[str]) -> None:
        """
        Note - all tags must be present in this filter.

        Not just any individual one.
        """
        self.__set_all('tags', all_)

    @property
    def __flow_run_filter(self) -> Optional[FlowRunFilter]:
        kwargs = {
            k: v
            for k, v in self.__filters.items()
            if not k.startswith('_')
        }
        if not kwargs:
            return None
        return self.__flow_run_class(**kwargs)

    @property
    def __deployment_filter(self) -> Optional[DeploymentFilter]:
        kwargs = self.__pull_out_filters('_deployment_')
        if not kwargs:
            return None
        return self.__deployment_class(**kwargs)

    @property
    def __flow_filter(self) -> Optional[FlowFilter]:
        kwargs = self.__pull_out_filters('_flow_')
        if not kwargs:
            return None
        return self.__flow_class(**kwargs)

    def __pull_out_filters(
        self,
        initial: str
    ) -> Optional[_FilterKwargsDict]:

        start = len(initial)

        return {
            k[start:]: v
            for k, v in self.__filters.items()
            if k.startswith(initial)
        }

    def __set_any(
        self,
        name: str,
        any_: list[str]
    ) -> None:

        self.__set(name, any_=any_)

    def __set_all(
        self,
        name: str,
        all_: list[str]
    ) -> None:

        self.__set(name, all_=all_)

    def __set(self, name: str, **kwargs) -> None:
        stripped = name.lstrip('_')
        class_ = self.__class_map[stripped]
        self.__filters[name] = class_(**kwargs)


class PrefectFilter:
    """
    Converts a `DataSourceFilter` instance to a form
    that `PrefectDataSource` can understand.
    """

    __VALID_KEYS = (
        'id',
        'idempotency_key',
        'name',
        'state',
        'tags',
        'flow_name',
        'deployment_name'
    )

    def __init__(
        self,
        builder: PrefectFilterBuilder
    ) -> None:

        self.__builder = builder

    def to_kwargs(
        self,
        filters: DataSourceFilter
    ) -> _FilterKwargsDict:
        """Returns filter kwargs to add to get operations"""

        self.__validate_filter_roots(filters)

        self.__add_all_exact(filters)
        self.__add_all_in_list(filters)

        return self.__builder.kwargs

    def __add_all_exact(
        self,
        filters: DataSourceFilter
    ) -> None:

        if not filters.exact:
            return

        for k, v in filters.exact.items():
            self.__add_exact(k, v)

    def __add_all_in_list(
        self,
        filters: DataSourceFilter
    ) -> None:

        if not filters.in_list:
            return

        for k, v in filters.in_list.items():
            self.__add_in_list(k, v)

    def __add_exact(self, k: str, v: Any) -> None:
        self.__check_valid_key(k)

        if k == 'tags':
            return self.__add_exact_tags(v)

        if isinstance(v, list):
            self.__raise_bad_list_exact(k)

        self.__get_builder_method(k)([v])

    def __add_in_list(self, k: str, v: list[str]) -> None:
        self.__check_valid_key(k)

        if k == 'tags':
            self.__in_list_error_tags()

        self.__get_builder_method(k)(v)

    def __add_exact_tags(
        self,
        v: Union[str, list[str]]
    ) -> None:

        if isinstance(v, str):
            return self.__builder.tags_all([v])

        self.__builder.tags_all(v)

    def __in_list_error_tags(self) -> None:
        raise DataSourceError(
            title='Bad Filter',
            detail='Cannot use `in_list` against tags',
            status_code=400
        )

    def __raise_bad_list_exact(self, k: str) -> None:
        raise DataSourceError(
            title='Bad Filter',
            detail=f'Cannot provide a list to exact on {k}.',
            status_code=400
        )

    def __validate_filter_roots(
        self,
        filters: DataSourceFilter
    ) -> None:

        if filters.contains or filters.range:
            raise DataSourceError(
                title='Bad Filter',
                detail='Can only use `exact` and `in_list` terms',
                status_code=400
            )

    def __check_valid_key(self, k: str) -> None:
        if k not in self.__VALID_KEYS:
            raise DataSourceError(
                title='Bad Filter',
                detail='Unknown filter key',
                status_code=400
            )

    def __get_builder_method(
        self,
        k: str
    ) -> Callable:

        if k == 'id':
            return self.__builder.id_

        return getattr(self.__builder, k)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Any, Iterable, Optional

from dateutil.parser import parse as dateutil_parse

from ..core import DataObject

if typing.TYPE_CHECKING:
    from ..core import DataSource


CopoApiResource = dict[str, Any]
CopoApiDoc = dict[str, list[CopoApiResource]]


class Parser(ABC):
    """
    Parses COPO API transfer resource `dict`s to `DataObject`
    instances
    """

    def parse_iterable(
        self,
        transfers: Iterable[CopoApiResource]
    ) -> Iterable[DataObject]:
        """
        Parses an `Iterable` of COPO API transfer resources
        """

        return (
            self.parse(t) for t in transfers
        )

    @abstractmethod
    def parse(
        self,
        transfer: CopoApiResource
    ) -> DataObject:
        """
        Parses an individual COPO transfer resource to a
        `DataObject` instance
        """


class DefaultParser(Parser):

    def __init__(self, data_source_dict: dict[str, DataSource]) -> None:
        self.__dict = data_source_dict

    def parse(
            self,
            transfer: CopoApiResource) -> DataObject:
        object_type = transfer.pop('tolsdk-type', 'sample')
        object_id = transfer.pop('copo_id')
        ds = self.__get_data_source(object_type)
        raw_attributes = transfer
        attributes = self.__convert_attributes(object_type, raw_attributes)

        return ds.data_object_factory(
            object_type,
            id_=object_id,
            attributes=attributes
        )

    def __get_data_source(self, type_: str) -> DataSource:
        return self.__dict[type_]

    def __convert_attributes(
        self,
        type_: str,
        attributes: Optional[dict[str, Any]]
    ) -> dict[str, Any]:

        if not attributes:
            return {}
        ds = self.__get_data_source(type_)
        attribute_types = ds.attribute_types.get(
            type_,
            {}
        )

        datetime_keys = self.__get_datetime_keys(type_)

        return {
            k: (
                dateutil_parse(v)
                if k in datetime_keys and v is not None and v != ''
                else v
            )
            for k, v in attributes.items()
            if k in attribute_types
        }

    def __get_datetime_keys(self, type_: str) -> list[str]:
        ds = self.__get_data_source(type_)
        attribute_types = ds.attribute_types.get(
            type_,
            {}
        )

        return [
            k for k, v in attribute_types.items()
            if self.__value_is_datetime(v)
        ]

    def __value_is_datetime(self, __v: str) -> bool:
        lower_ = __v.lower()

        return 'date' in lower_ or 'time' in lower_



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .allowed_keys import AllowedKeysValidator  # noqa
from .allowed_values import AllowedValuesValidator  # noqa
from .allowed_values_from_datasource import AllowedValuesFromDataSourceValidator  # noqa
from .assert_on_condition import AssertOnConditionValidator  # noqa
from .converter_and_validate import ConverterAndValidateValidator  # noqa
from .ena_checklist import EnaChecklistValidator # noqa
from .mutually_exclusive import MutuallyExclusiveValidator  # noqa
from .ena_submittable import EnaSubmittableValidator  # noqa
from .regex import Regex, RegexValidator  # noqa
from .regex_by_value import RegexByValueValidator  # noqa
from .specimens_have_same_taxon import SpecimensHaveSameTaxonValidator # noqa
from .sts_fields import StsFieldsValidator  # noqa
from .tolid import TolidValidator  # noqa
from .types import TypesValidator  # noqa
from .unique_values import UniqueValuesValidator  # noqa
from .unique_whole_organisms import UniqueWholeOrganismsValidator  # noqa
from .interfaces import Condition  # noqa
from .min_one_valid_value import MinOneValidValueValidator  # noqa
from .value_check import ValueCheckValidator  # noqa
from .branching import BranchingValidator  # noqa
from .unique_value_check import UniqueValueCheckValidator  # noqa
from .date_sorting import DateSortingValidator  # noqa
from .taxon_matches_goat import TaxonMatchesGoatValidator  # noqa
from .json_schema import JsonSchemaValidator  # noqa



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Iterable, Optional, Tuple

from ._filterable import _Filterable

if typing.TYPE_CHECKING:
    from ..data_object import DataObject
    from ..datasource_filter import DataSourceFilter
    from ..requested_fields import ReqFieldsTree
    from ..session import OperableSession


class PageGetter(_Filterable, ABC):
    """
    Gets pages of DataObject instances.
    """

    @abstractmethod
    def get_list_page(
        self,
        object_type: str,
        page_number: int,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        sort_by: Optional[str] = None,
        session: Optional[OperableSession] = None,
        requested_fields: list[str] | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Tuple[Iterable[DataObject], int]:
        """
        For a specified object_type, of the given page_size
        and page_number (starting from 1), returns a tuple of:

        - An Iterable of DataObject instances
        - The total number of DataObjects that matches the filter
        """



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Any, Optional

from ._filterable import _Filterable

if typing.TYPE_CHECKING:
    from ..datasource_filter import DataSourceFilter
    from ..session import OperableSession


class Aggregator(_Filterable, ABC):
    """
    Gets aggregations according to the Elastic aggregation specification
    """

    @abstractmethod
    def get_aggregations(
        self,
        object_type: str,
        aggregations: dict[str, Any],
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None
    ) -> dict[str, Any]:
        """
        Gets aggregations according to the Elastic aggregation specification
        """



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from ..benchling import (
    BenchlingDataSource
)
from ..core import (
    core_data_object
)


def benchling(**kwargs) -> BenchlingDataSource:
    benchling = BenchlingDataSource(
        {
            'api_key': os.getenv('BENCHLING_API_KEY'),
            'url': os.getenv('BENCHLING_URL'),
            'registry_id': os.getenv('BENCHLING_REGISTRY_ID'),
            'project_id': os.getenv('BENCHLING_PROJECT_ID')
        }
    )
    core_data_object(benchling)

    return benchling



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import inspect
import re
import urllib.parse
from typing import Dict, Iterable, List

import mysql.connector

from ..core import DataObject, DataSource, DataSourceError, DataSourceFilter
from ..core.operator import DetailGetter, ListGetter


class MlwhDataSource(DataSource, DetailGetter, ListGetter):
    def __init__(self, config: Dict):
        # uri
        super().__init__(config, expected=['uri'])
        self._initialise_mlwh()

    def _initialise_mlwh(self):
        # Connect to MLWH
        mlwh_settings = urllib.parse.urlparse(self.uri)
        self.mlwh = mysql.connector.connect(
            user=mlwh_settings.username,
            password=urllib.parse.unquote(mlwh_settings.password),
            host=mlwh_settings.hostname,
            port=mlwh_settings.port,
            database=mlwh_settings.path[1:],
        )

    def _columns_string_from_mappings(self, mappings):
        return '\n              , '.join(
            f'{col} AS {alias}' for alias, col in mappings.items()
        )

    def _get_column_mappings_illumina(self):
        return {
            'id': (
                # Trim file suffix, i.e. ".cram"
                'REGEXP_REPLACE(irods.irods_data_relative_path'
                ", '\\.[[:alnum:]]+$', '')"
            ),
            'study_id': 'CONVERT(study.id_study_lims, SIGNED)',
            'sample_ref': 'sample.name',
            'supplier_name': 'sample.supplier_name',
            'tolid': 'sample.public_name',
            'biosample_accession': 'sample.accession_number',
            'biospecimen_accession': 'sample.donor_id',
            'scientific_name': 'sample.common_name',
            'taxon_id': 'sample.taxon_id',
            'platform_type': "'Illumina'",
            'instrument_model': 'run_lane_metrics.instrument_model',
            'instrument_name': 'run_lane_metrics.instrument_name',
            'element':
                'COALESCE(REGEXP_SUBSTR('
                'irods.irods_data_relative_path'
                ", '(?<=_)[^_]+(?=#)')"
                ', run_lane_metrics.position)',
            'pipeline_id_lims': 'flowcell.pipeline_id_lims',
            'run_id': 'COALESCE(REGEXP_SUBSTR('
                'irods.irods_data_relative_path, '
                "'^[^#\\.]+'"
                '), CONVERT(product_metrics.id_run, CHAR))',
            'tag_index': 'CONVERT(product_metrics.tag_index, CHAR)',
            'run_complete': 'run_lane_metrics.run_complete',
            'lims_qc': (
                'IF(product_metrics.qc IS NULL, NULL,'
                " IF(product_metrics.qc = 1, 'pass', 'fail'))"
            ),
            'cost_code': 'flowcell.cost_code',
            'qc_date': 'run_lane_metrics.qc_complete',
            'tag1_id': 'flowcell.tag_identifier',
            'tag2_id': 'flowcell.tag2_identifier',
            'library_id': 'flowcell.id_library_lims',
            'irods_path': 'irods.irods_root_collection',
            'irods_file': 'irods.irods_data_relative_path',
        }

    def _get_illumina_query(self, clause: str):
        col_string = self._columns_string_from_mappings(
            self._get_column_mappings_illumina()
        )

        return inspect.cleandoc(
            f"""
            SELECT {col_string}
            FROM study
            JOIN iseq_flowcell AS flowcell
              ON study.id_study_tmp = flowcell.id_study_tmp
            JOIN sample
              ON flowcell.id_sample_tmp = sample.id_sample_tmp
            JOIN iseq_product_metrics AS component_metrics
              ON flowcell.id_iseq_flowcell_tmp
                 = component_metrics.id_iseq_flowcell_tmp
            JOIN iseq_run_lane_metrics AS run_lane_metrics
              ON component_metrics.id_run = run_lane_metrics.id_run
              AND component_metrics.position = run_lane_metrics.position
            JOIN iseq_product_components AS components
              ON component_metrics.id_iseq_pr_metrics_tmp
                 = components.id_iseq_pr_component_tmp
              AND components.component_index = 1
            JOIN iseq_product_metrics AS product_metrics
              ON components.id_iseq_pr_tmp
                 = product_metrics.id_iseq_pr_metrics_tmp
            JOIN seq_product_irods_locations AS irods
              ON product_metrics.id_iseq_product = irods.id_product
            WHERE {clause}
              AND run_lane_metrics.qc_complete IS NOT NULL
              AND study.id_lims = 'SQSCP'
            ORDER BY run_lane_metrics.id_run
            """,
        )

    def _get_column_mappings_pacbio(self):
        return {
            'id': 'well_metrics.movie_name',
            'study_id': 'CONVERT(study.id_study_lims, SIGNED)',
            'sample_ref': 'sample.name',
            'supplier_name': 'sample.supplier_name',
            'tolid': 'sample.public_name',
            'biosample_accession': 'sample.accession_number',
            'biospecimen_accession': 'sample.donor_id',
            'scientific_name': 'sample.common_name',
            'taxon_id': 'sample.taxon_id',
            'platform_type': "'PacBio'",
            'instrument_model': (
                "REGEXP_REPLACE(instrument_type, '^Sequel2', 'Sequel II')"
            ),
            'instrument_name': "CONCAT('m', LOWER(instrument_name))",
            'pipeline_id_lims': 'run.pipeline_id_lims',
            'run_id': 'well_metrics.movie_name',
            'lims_run_id': 'well_metrics.pac_bio_run_name',
            'element': 'IF(well_metrics.plate_number IS NULL'
                       ', well_metrics.well_label'
                       ', CONCAT(well_metrics.well_label'
                       ", '.'"
                       ', well_metrics.plate_number))',
            'run_start': 'well_metrics.run_start',
            'run_complete': 'well_metrics.run_complete',
            'plex_count': 'plex_agg.plex_count',
            'lims_qc': (
                'IF(well_metrics.qc_seq IS NULL, NULL,'
                " IF(well_metrics.qc_seq = 1, 'pass', 'fail'))"
            ),
            'qc_date': 'well_metrics.qc_seq_date',
            'qc_seq_state': 'well_metrics.qc_seq_state',
            'qc_seq_state_is_final': (
                'IF(well_metrics.qc_seq_state_is_final IS NULL, NULL,'
                ' IF(well_metrics.qc_seq_state_is_final = 1, true, false))'
            ),
            'cost_code': 'run.cost_code',
            'pac_bio_library_tube_name': 'run.pac_bio_library_tube_name',
            'tag1_id': 'run.tag_identifier',
            'tag2_id': 'run.tag2_identifier',
            'library_id': 'run.pac_bio_library_tube_name',
            # Extra fields for tolqc.pacbio_run_metrics:
            'movie_minutes': 'well_metrics.movie_minutes',
            'binding_kit': 'well_metrics.binding_kit',
            'sequencing_kit': 'well_metrics.sequencing_kit',
            'sequencing_kit_lot_number': 'well_metrics.sequencing_kit_lot_number',
            'cell_lot_number': 'well_metrics.cell_lot_number',
            'include_kinetics': (
                'IF(well_metrics.include_kinetics IS NULL, NULL,'
                " IF(well_metrics.include_kinetics = 1, 'true', 'false'))"
            ),
            'loading_conc': 'well_metrics.loading_conc',
            'control_num_reads': 'well_metrics.control_num_reads',
            'control_read_length_mean': 'well_metrics.control_read_length_mean',
            'control_concordance_mean': 'well_metrics.control_concordance_mean',
            'control_concordance_mode': 'well_metrics.control_concordance_mode',
            'local_base_rate': 'well_metrics.local_base_rate',
            'polymerase_read_bases': 'well_metrics.polymerase_read_bases',
            'polymerase_num_reads': 'well_metrics.polymerase_num_reads',
            'polymerase_read_length_mean': 'well_metrics.polymerase_read_length_mean',
            'polymerase_read_length_n50': 'well_metrics.polymerase_read_length_n50',
            'insert_length_mean': 'well_metrics.insert_length_mean',
            'insert_length_n50': 'well_metrics.insert_length_n50',
            'unique_molecular_bases': 'well_metrics.unique_molecular_bases',
            'productive_zmws_num': 'well_metrics.productive_zmws_num',
            'p0_num': """
                (well_metrics.p0_num * 100) /
                NULLIF((well_metrics.p0_num + well_metrics.p1_num + well_metrics.p2_num), 0)
            """,
            'p1_num': """
                (well_metrics.p1_num * 100) /
                NULLIF((well_metrics.p0_num + well_metrics.p1_num + well_metrics.p2_num), 0)
            """,
            'p2_num': """
                (well_metrics.p2_num * 100) /
                NULLIF((well_metrics.p0_num + well_metrics.p1_num + well_metrics.p2_num), 0)
            """,
            'adapter_dimer_percent': 'well_metrics.adapter_dimer_percent',
            'short_insert_percent': 'well_metrics.short_insert_percent',
            'well_hifi_read_bases': 'well_metrics.hifi_read_bases',
            'product_hifi_read_bases': 'product_metrics.hifi_read_bases',
            'hifi_num_reads': 'well_metrics.hifi_num_reads',
            'hifi_read_length_mean': 'well_metrics.hifi_read_length_mean',
            'hifi_read_quality_median': 'well_metrics.hifi_read_quality_median',
            'hifi_number_passes_mean': 'well_metrics.hifi_number_passes_mean',
            'hifi_low_quality_read_bases': 'well_metrics.hifi_low_quality_read_bases',
            'hifi_low_quality_num_reads': 'well_metrics.hifi_low_quality_num_reads',
            'hifi_low_quality_read_length_mean':
                'well_metrics.hifi_low_quality_read_length_mean',
            'hifi_low_quality_read_quality_median':
                'well_metrics.hifi_low_quality_read_quality_median',
            'hifi_barcoded_reads': 'well_metrics.hifi_barcoded_reads',
            'hifi_bases_in_barcoded_reads': 'well_metrics.hifi_bases_in_barcoded_reads',
            'irods_path': 'irods.irods_root_collection',
            'irods_file': 'irods.irods_data_relative_path',
        }

    def _get_pacbio_query(self, clause: str):
        col_string = self._columns_string_from_mappings(
            self._get_column_mappings_pacbio()
        )
        return inspect.cleandoc(
            f"""
            WITH plex_agg AS (
                SELECT rwm.id_pac_bio_rw_metrics_tmp, COUNT(*) plex_count
                FROM pac_bio_run_well_metrics rwm
                JOIN pac_bio_product_metrics pm
                USING (id_pac_bio_rw_metrics_tmp)
                GROUP BY rwm.id_pac_bio_rw_metrics_tmp
            )
            SELECT {col_string}
            FROM study
            JOIN pac_bio_run AS run
                ON study.id_study_tmp = run.id_study_tmp
            JOIN sample
                ON run.id_sample_tmp = sample.id_sample_tmp
            JOIN pac_bio_product_metrics AS product_metrics
                ON run.id_pac_bio_tmp = product_metrics.id_pac_bio_tmp
            JOIN pac_bio_run_well_metrics AS well_metrics
                ON product_metrics.id_pac_bio_rw_metrics_tmp
                = well_metrics.id_pac_bio_rw_metrics_tmp
            JOIN plex_agg
                ON well_metrics.id_pac_bio_rw_metrics_tmp
                = plex_agg.id_pac_bio_rw_metrics_tmp
            LEFT JOIN seq_product_irods_locations AS irods
                ON product_metrics.id_pac_bio_product = irods.id_product
            WHERE {clause}
              AND well_metrics.movie_name IS NOT NULL
              AND study.id_lims = 'SQSCP'
            ORDER BY well_metrics.pac_bio_run_name
            """
        )

    def _get_column_mappings_sequencing_request(self):
        return {
            'sample_ref': 'sample.friendly_name',
            'supplier_name': 'mlwh_sample.supplier_name',
            'accession_number': 'mlwh_sample.accession_number',
            'public_name': 'mlwh_sample.public_name',
            'donor_id': 'mlwh_sample.donor_id',
            'taxon_id': 'mlwh_sample.taxon_id',
            'common_name': 'mlwh_sample.common_name',
            'description': 'mlwh_sample.description',
            'study_id': 'mlwh_study.id_study_lims',
            # 'study_uuid': 'study.uuid',
            'order_date': "DATE_FORMAT(MIN(events.created_at), '%Y-%m-%dT%H:%i:%s')",
        }

    def _get_sequencing_request_query(self, clause: str):
        col_string = self._columns_string_from_mappings(
            self._get_column_mappings_sequencing_request()
        )
        return inspect.cleandoc(
            f"""
            SELECT {col_string}
            FROM mlwh_events.events
            JOIN mlwh_events.event_types
              ON event_types.id = events.event_type_id
            JOIN mlwh_events.roles AS sample_roles
              ON events.id = sample_roles.event_id
              AND sample_roles.role_type_id = 6
            JOIN mlwh_events.subjects AS sample
              ON sample_roles.subject_id = sample.id
            JOIN mlwh_events.roles AS study_roles
              ON events.id = study_roles.event_id
              AND study_roles.role_type_id = 2
            JOIN mlwh_events.subjects AS study
              ON study_roles.subject_id = study.id
            LEFT JOIN mlwarehouse.sample mlwh_sample
              ON sample.friendly_name = mlwh_sample.name
              AND mlwh_sample.id_lims = 'SQSCP'
            JOIN mlwarehouse.study AS mlwh_study
              ON UNHEX(
                REPLACE(mlwh_study.uuid_study_lims, '-', '')
              ) = study.uuid
              AND mlwh_study.id_lims = 'SQSCP'
            WHERE {clause}
              AND event_types.key = 'order_made'
            GROUP BY sample.friendly_name
            """
        )

    def _get_column_mappings_long_read_qc_result(self):
        return {
            'id': 'id_long_read_qc_result_tmp',
            'labware_barcode': 'labware_barcode',
            'sample_id': 'sample_id',
            'assay_type': 'assay_type_key',
            'units': 'units',
            'value': 'CAST(value AS DECIMAL(20, 2))',
            'recorded_at': 'recorded_at',
            'qc_status': 'qc_status',
            'qc_status_decision_by': 'qc_status_decision_by',
        }

    def _get_long_read_qc_result_query(self, clause: str):
        col_string = self._columns_string_from_mappings(
            self._get_column_mappings_long_read_qc_result()
        )
        return inspect.cleandoc(
            f"""
            SELECT {col_string}
            FROM long_read_qc_result
            WHERE {clause}
            ORDER BY id_long_read_qc_result_tmp, recorded_at
            """
        )

    def _get_column_mappings_study(self):
        return {
            'id': 'id_study_tmp',
            'id_lims': 'id_lims',
            'uuid_study_lims': 'uuid_study_lims',
            'id_study_lims': 'id_study_lims',
            'last_updated': 'last_updated',
            'recorded_at': 'recorded_at',
            'deleted_at': 'deleted_at',
            'created': 'created',
            'name': 'name',
            'reference_genome': 'reference_genome',
            'ethically_approved': 'ethically_approved',
            'faculty_sponsor': 'faculty_sponsor',
            'state': 'state',
            'study_type': 'study_type',
            'abstract': 'abstract',
            'abbreviation': 'abbreviation',
            'accession_number': 'accession_number',
            'description': 'description',
            'contains_human_dna': 'contains_human_dna',
            'contaminated_human_dna': 'contaminated_human_dna',
            'data_release_strategy': 'data_release_strategy',
            'data_release_sort_of_study': 'data_release_sort_of_study',
            'ena_project_id': 'ena_project_id',
            'study_title': 'study_title',
            'study_visibility': 'study_visibility',
            'ega_dac_accession_number': 'ega_dac_accession_number',
            'array_express_accession_number': 'array_express_accession_number',
            'ega_policy_accession_number': 'ega_policy_accession_number',
            'data_release_timing': 'data_release_timing',
            'data_release_delay_period': 'data_release_delay_period',
            'data_release_delay_reason': 'data_release_delay_reason',
            'aligned': 'aligned',
            'separate_y_chromosome_data': 'separate_y_chromosome_data',
            'prelim_id': 'prelim_id',
            'hmdmc_number': 'hmdmc_number',
            'data_destination': 'data_destination',
            's3_email_list': 's3_email_list',
            'data_deletion_period': 'data_deletion_period',
            'contaminated_human_data_access_group':
                'contaminated_human_data_access_group',
            'programme': 'programme',
            'ebi_library_strategy': 'ebi_library_strategy',
            'ebi_library_source': 'ebi_library_source',
            'ebi_library_selection': 'ebi_library_selection',
        }

    def _get_study_query(self, clause: str):
        col_string = self._columns_string_from_mappings(
            self._get_column_mappings_study()
        )
        return inspect.cleandoc(
            f"""
            SELECT {col_string}
            FROM study
            WHERE {clause}
            ORDER BY id_study_tmp
            """
        )

    def _get_column_mappings_sequencing_request_volume(self):
        return {
            'id': 'aliquot.sample_name',
            'original_volume': 'aliquot.volume',
            'insert_size': 'aliquot.insert_size',
            'concentration': 'aliquot.concentration',
            'source_barcode': 'aliquot.source_barcode',
            'volume_remaining': 'aliquot.volume - COALESCE(derived_amounts.amount, 0)'
        }

    def _get_sequencing_request_volume_query(self, clause: str):
        col_string = self._columns_string_from_mappings(
            self._get_column_mappings_sequencing_request_volume()
        )

        return inspect.cleandoc(
            f"""
            -- derived
            WITH derived_amounts AS (
                SELECT sample_name, SUM(volume) as amount
                FROM aliquot
                WHERE aliquot_type = 'derived'
                AND source_type = 'library'
                GROUP BY sample_name
            ),
            latest_primary_aliquots AS (
                SELECT sample_name, MAX(last_updated) AS latest_updated_at
                FROM aliquot
                WHERE aliquot_type = 'primary'
                AND source_type = 'library'
                GROUP BY sample_name
            )
            -- main SELECT
            SELECT {col_string}
            FROM aliquot
            LEFT JOIN derived_amounts on aliquot.sample_name = derived_amounts.sample_name
            INNER JOIN latest_primary_aliquots
                ON aliquot.sample_name = latest_primary_aliquots.sample_name
                AND aliquot.last_updated = latest_primary_aliquots.latest_updated_at
            WHERE {clause}
            AND aliquot.aliquot_type='primary'
            AND aliquot.source_type = 'library'
            """
        )

    def _get_id_from_row(self, row: Dict) -> str:
        data_id = row.get('id')
        if row.get('platform_type', '').lower() == 'pacbio':
            tag1 = self.__trimmed_tag(row.get('tag1_id'))
            tag2 = self.__trimmed_tag(row.get('tag2_id'))
            if tag2:
                return f'{data_id}#{tag1}#{tag2}'
            elif tag1:
                return f'{data_id}#{tag1}'
        return data_id

    def __trimmed_tag(self, tag):
        if tag is None:
            return tag
        if m := re.match(r'bc(\d{4,})', tag):
            return m.group(1)
        return tag

    def _format_mlwh_row(self, object_type: str, row: Dict):
        data_id = self._get_id_from_row(row)
        return self.data_object_factory(
            object_type,
            id_=data_id,
            attributes={
                k: v
                for k, v in row.items()
                if k != 'id'
            }
        )

    def _join(self, values: List) -> str:
        return "','".join([str(s) for s in values])

    def _conditions_string(self, platform_type: str, in_list: Dict):
        if not in_list:
            return '1=1'  # Something to go with the where clause
        sql_conditions = []
        if platform_type.lower() == 'illumina':
            mappings = self._get_column_mappings_illumina()
        if platform_type.lower() == 'pacbio':
            mappings = self._get_column_mappings_pacbio()
        if platform_type.lower() == 'sequencing_request':
            mappings = self._get_column_mappings_sequencing_request()
        if platform_type.lower() == 'sequencing_request_volume':
            mappings = self._get_column_mappings_sequencing_request_volume()
        if platform_type.lower() == 'study':
            mappings = self._get_column_mappings_study()
        for k, v in in_list.items():
            mapped_k = mappings[k]
            sql_conditions.append(f"{mapped_k} IN ('{self._join(v)}')")
        sql_conditions_string = ' AND '.join(sql_conditions)
        return sql_conditions_string

    def _execute_query(self, query, object_type):
        cur_mlwh = self.mlwh.cursor(dictionary=True)
        cur_mlwh.execute(query)
        for row in cur_mlwh:
            yield self._format_mlwh_row(object_type, row)

    def __get_in_lists(self, f: DataSourceFilter):
        ret = {}
        for a in f.and_:
            if 'in_list' in f.and_[a] and 'value' in f.and_[a]['in_list']:
                ret[a] = f.and_[a]['in_list']['value']
        return ret

    def get_list(
        self,
        object_type: str,
        object_filters: DataSourceFilter = None,
        **kwargs,
    ) -> Iterable[DataObject]:
        # Sort out the conditions
        if object_type == 'run_data':
            if (
                object_filters is None
                or not isinstance(object_filters.and_, dict)
                or 'platform_type' not in object_filters.and_
                or 'eq' not in object_filters.and_['platform_type']
                or 'value' not in object_filters.and_['platform_type']['eq']
            ):
                raise DataSourceError(
                    'Filter must contain platform_type exact filter and list of study_ids'
                )
            if object_filters.and_['platform_type']['eq']['value'].lower() == 'illumina':
                sql_conditions = self._conditions_string(
                    'illumina', self.__get_in_lists(object_filters)
                )
                query = self._get_illumina_query(sql_conditions)
                return self._execute_query(query, 'run_data')
            elif object_filters.and_['platform_type']['eq']['value'].lower() == 'pacbio':
                sql_conditions = self._conditions_string(
                    'pacbio',
                    self.__get_in_lists(object_filters)
                )
                query = self._get_pacbio_query(sql_conditions)
                return self._execute_query(query, 'run_data')
        elif object_type == 'sequencing_request':
            sql_conditions = self._conditions_string(
                'sequencing_request',
                self.__get_in_lists(object_filters)
            )
            query = self._get_sequencing_request_query(sql_conditions)
            return self._execute_query(query, 'sequencing_request')
        elif object_type == 'long_read_qc_result':
            sql_conditions = self._conditions_string(
                'long_read_qc_result', None  # No filters
            )
            query = self._get_long_read_qc_result_query(sql_conditions)
            return self._execute_query(query, 'long_read_qc_result')
        elif object_type == 'study':
            sql_conditions = self._conditions_string(
                'study',
                self.__get_in_lists(object_filters)
            )
            query = self._get_study_query(sql_conditions)
            return self._execute_query(query, 'study')
        else:
            raise DataSourceError(
                'Only objects of type long_read_qc_result, run_data, study or '
                'sequencing_request are supported'
            )

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        **kwargs,
    ) -> Iterable[DataObject]:
        # Sort out the conditions
        if object_type == 'sequencing_request_volume':
            sql_conditions = self._conditions_string(
                'sequencing_request_volume', {'id': object_ids}
            )
            query = self._get_sequencing_request_volume_query(sql_conditions)
            return self._execute_query(query, 'sequencing_request_volume')
        else:
            raise DataSourceError(
                'Only objects of type sequencing_request_volume are supported'
            )

    @property
    def supported_types(self) -> List[str]:
        return ['long_read_qc_result', 'sequencing_request', 'run_data',
                'sequencing_request_volume', 'study']



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os
import uuid
from tempfile import NamedTemporaryFile
from typing import Any

from flask import Blueprint, request, send_file

from .blueprint import (
    custom_blueprint,
)
from ..services.s3_client import S3Client


ALLOWED_EXTENSIONS: set[str] = {'csv', 'json', 'xlsx'}


def allowed_file(filename: str) -> bool:
    return '.' in filename and filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS


def set_filename(filename: str) -> str:
    environment = os.getenv('ENVIRONMENT', 'dev')
    unique_id = str(uuid.uuid4())[:8]
    return f'{environment}_{unique_id}_{filename}'


def get_original_filename(s3_filename: str) -> str:
    return s3_filename.split('_', maxsplit=2)[2]


def data_upload_blueprint(
    url_prefix: str = '/pipeline/data_upload',
) -> Blueprint:

    data_upload_blueprint = custom_blueprint(
        name='data_upload',
        url_prefix=url_prefix
    )

    @data_upload_blueprint.route('/upload', methods=['POST'])
    def upload_file() -> tuple[dict[str, str], int]:
        file = request.files['file']
        s3_bucket: str = request.form.get('s3_bucket')

        if not file:
            return {'error': 'No file provided'}, 400

        if not allowed_file(file.filename):
            return {'error': 'File type not allowed'}, 400

        if not s3_bucket:
            return {'error': 'S3 bucket not specified'}, 400

        try:
            s3_client = S3Client()
            new_filename = set_filename(file.filename)

            with NamedTemporaryFile() as temp_file:
                file.save(temp_file.name)
                s3_client.put_object(s3_bucket, new_filename, temp_file.name)

            return {'message': 'File uploaded successfully', 'file_name': new_filename}, 200

        except Exception as e:
            return {'error': f'Failed to upload file: {str(e)}'}, 500

    @data_upload_blueprint.route('/download', methods=['POST'])
    def download_file() -> tuple[dict[str, str], int]:
        body: dict[str, Any] = request.json.get('data', {})

        if 's3_bucket' not in body or 'file_name' not in body:
            return {'error': 'S3 bucket or file name not specified'}, 400

        bucket_name = body['s3_bucket']
        file_name = body['file_name']

        try:
            s3_client = S3Client()

            with NamedTemporaryFile() as temp_file:
                s3_client.get_object(bucket_name, file_name, temp_file.name)

                original = get_original_filename(file_name)

                return send_file(
                    temp_file.name,
                    as_attachment=True,
                    download_name=original,
                    mimetype='application/octet-stream'
                )
        except Exception as e:
            return {'error': f'Failed to download file: {str(e)}'}, 500

    return data_upload_blueprint



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Dict, Iterable, Iterator, List, Optional

from cachetools import LFUCache

from .operator import Relational

if typing.TYPE_CHECKING:
    from .data_object import DataObject


RelationshipDump = Dict[str, Dict[str, str]]


@dataclass
class RelationshipConfig:
    """
    Describes the configuration of relationships on a specific
    type of DataObject.

    The keys of each Optional[Dict] are the names of a relationship,
    and the values are the type of DataObject instances to which they
    are directed. For foreign keys, the values are the name of the
    attribute in this class (to_one) or the foreign class (to_many)
    that implements the relationship.
    """

    to_one: Optional[Dict[str, str]] = None
    to_many: Optional[Dict[str, str | List[str]]] = None
    foreign_keys: Optional[Dict[str, str]] = None

    def to_dict(self) -> Optional[RelationshipDump]:
        if self.empty:
            return None
        return self.__build_dict()

    @property
    def empty(self) -> bool:
        return self.to_one is None and self.to_many is None

    def __build_dict(self) -> RelationshipDump:
        __dict = {}
        if self.to_one is not None:
            __dict['one'] = self.to_one
        if self.to_many is not None:
            __dict['many'] = self.to_many
        if self.foreign_keys is not None:
            __dict['foreign_keys'] = self.foreign_keys
        return __dict

    def __str__(self) -> str:
        return str(self.to_dict())


class ToManyDict(Mapping):
    """A Dict that loads items lazily and memoizes the result"""

    def __init__(self, source: DataObject) -> None:
        self.__dict = LFUCache(100000)
        self.__source = source
        self.__host: Relational = source._host
        self.__keys = self.__get_keys()

    def __getitem__(self, __k: str) -> Iterable[DataObject]:
        if __k not in self.__keys:
            raise KeyError()
        if __k not in self.__dict:
            return self.__get_many_objects(__k)
        else:
            return self.__dict[__k]

    def __iter__(self) -> Iterator[str]:
        return iter(self.__keys)

    def __len__(self) -> int:
        return len(self.__keys)

    def __get_many_objects(self, __k: str) -> Iterable[DataObject]:
        many_objects = self.__host.get_to_many_relations(
            self.__source,
            __k
        )
        self.__dict[__k] = many_objects
        return many_objects

    def __get_keys(self) -> List[str]:
        config = self.__config
        if config is None or config.to_many is None:
            return []
        return list(config.to_many.keys())

    @property
    def __config(self) -> Optional[RelationshipConfig]:
        type_ = self.__source.type
        return self.__host.relationship_config.get(type_)


class ToOneDict(Mapping):
    """
    A dict that implements to-one relationships, and supports:

    - lazy loading
    - memoisation
    - overwriting

    This is only to be used by a CoreDataObject.
    """

    def __init__(self, source: DataObject) -> None:
        self.__dict = LFUCache(100000)
        self.__source = source
        self.__host: Relational = source._host
        self.__keys = self.__get_keys()

    def __getitem__(self, __k: str) -> Optional[DataObject]:
        if __k not in self.__keys:
            raise KeyError()
        set_to_one_objects = self.__source._to_one_objects
        if __k in set_to_one_objects:
            return set_to_one_objects[__k]
        return self.__get_or_cached(__k)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__keys)

    def __len__(self) -> int:
        return len(self.__keys)

    def __get_keys(self) -> List[str]:
        config = self.__config
        if config is None or config.to_one is None:
            return []
        return list(config.to_one.keys())

    def __get_one_object(self, __k: str) -> Optional[DataObject]:
        one_object = self.__host.get_to_one_relation(self.__source, __k)
        self.__dict[__k] = one_object
        return one_object

    def __get_or_cached(self, __k: str) -> Optional[DataObject]:
        if __k not in self.__dict:
            return self.__get_one_object(__k)
        else:
            return self.__dict[__k]

    @property
    def __config(self) -> Optional[RelationshipConfig]:
        type_ = self.__source.type
        return self.__host.relationship_config.get(type_)



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Any, Optional

from ..core import DataObject


class FlowRunObject(DataObject):
    """
    Contains fields relevant for a `Flow Run`
    in prefect.

    Used only for type hints.
    """

    flow_name: str
    deployment_name: str

    name: Optional[str]
    tags: Optional[list[str]]
    state: Optional[str]
    idempotency_key: Optional[str]
    parameters: Optional[dict[str, Any]]



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections.abc import Mapping
from typing import Callable, Iterator, Optional

from .client import LabwhereApiClient
from .converter import (
    LabwhereApiConverter
)
from .labwhere_datasource import (
    LabwhereConverterFactory,
    LabwhereDataSource
)
from .parser import DefaultParser
from ..core import DataSource


class _LabwhereDSDict(Mapping):
    def __init__(self, api_ds: LabwhereDataSource) -> None:
        self.__ds = api_ds

    def __getitem__(self, __k: str) -> LabwhereDataSource:
        if __k not in self.__ds.supported_types:
            raise KeyError()
        return self.__ds

    def __iter__(self) -> Iterator[str]:
        return iter(self.__ds.supported_types)

    def __len__(self) -> int:
        return len(self.__ds.supported_types)


class _ConverterFactory:
    """
    Manges the instantation of:

    - `LabwhereApiConverter`
    """

    def __init__(self) -> None:
        self.__data_source: Optional[DataSource] = None

    @property
    def data_source(self) -> Optional[DataSource]:
        return self.__data_source

    @data_source.setter
    def data_source(
        self,
        ds: DataSource
    ) -> None:

        self.__data_source = ds

    def labwhere_converter_factory(self) -> LabwhereConverterFactory:
        """
        Returns an instantiated `LabwhereApiConverter`.
        """

        parser = DefaultParser(self.__ds_dict)
        return LabwhereApiConverter(parser)

    @property
    def __ds_dict(self) -> dict[str, DataSource]:
        return _LabwhereDSDict(self.data_source)


def _get_client_factory(
    api_url: str
) -> Callable[[], LabwhereApiClient]:
    """
    A resonable default for creating
    a `JsonApiClient` instance
    """

    return lambda: LabwhereApiClient(
        api_url
    )


def create_labwhere_datasource(
    labwhere_url: str
) -> LabwhereDataSource:
    """
    Instantiates `LabwhereDataSource` using the given:

    - `labwhere_url`
    """

    client_factory = _get_client_factory(
        labwhere_url
    )
    manager = _ConverterFactory()

    labwhere_ds = LabwhereDataSource(
        client_factory,
        manager.labwhere_converter_factory
    )

    manager.data_source = labwhere_ds

    return labwhere_ds



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass

from tol.core import Validator
from tol.core.data_object import DataObject


class ValueCheckValidator(Validator):
    """
    Validates an incoming stream of `DataObject` instances.
    For each data object (sample) it checks if it is a SYMBIONT
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        field: str
        value: str

    __slots__ = ['__config']
    __config: Config

    def __init__(self, config: Config, **kwargs) -> None:
        super().__init__()
        self.__config = config

    def _validate_data_object(self, obj: DataObject) -> None:
        if obj.attributes.get(self.__config.field) != self.__config.value:
            self.add_error(
                object_id=obj.id,
                detail=(
                    f"Expected '{self.__config.value}' for field "
                    f"'{self.__config.field}', but got '{obj.attributes.get(self.__config.field)}'"
                ),
                field=self.__config.field,
            )



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Generic, TypeVar


In = TypeVar('In')
Out = TypeVar('Out')


class DataSourceParser(ABC, Generic[In, Out]):
    """
    Parses resources from the format `In` to the format `Out`
    """
    @abstractmethod
    def parse(self, transfer: In) -> Out | None:
        """
        Parses an individual transfer resource to a
        `DataObject` instance
        """
        raise NotImplementedError

    def parse_iterable(
        self, transfers: Iterable[In]
    ) -> Iterable[Out | None]:
        """
        Uses the subclass's implementation of `parse`
        to parse an iterable of transfer resources
        """
        return (self.parse(t) for t in transfers)



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

# THIS WILL BE DELETED AT SOME POINT. USE StsDataSource INSTEAD

import os

import requests


STS_URL = os.getenv('STS_URL')
STS_API_KEY = os.getenv('STS_API_KEY')
STS_API_PATH = os.getenv('STS_LEGACY_API_PATH')


def __override_method(method, relative_url, headers=None, **kwargs):
    if headers is None:
        new_headers = {
            'token': STS_API_KEY
        }
    else:
        new_headers = {
            'token': STS_API_KEY,
            **headers
        }
    return method(
        f'{STS_URL}{STS_API_PATH}/{relative_url}',
        headers=new_headers,
        **kwargs
    )


def get(relative_url, **kwargs):
    return __override_method(
        requests.get,
        relative_url,
        **kwargs
    )


def post(relative_url, **kwargs):
    return __override_method(
        requests.post,
        relative_url,
        **kwargs
    )


def put(relative_url, **kwargs):
    return __override_method(
        requests.put,
        relative_url,
        **kwargs
    )


def patch(relative_url, **kwargs):
    return __override_method(
        requests.patch,
        relative_url,
        **kwargs
    )


def delete(relative_url, **kwargs):
    return __override_method(
        requests.delete,
        relative_url,
        **kwargs
    )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


class AuthError(Exception):
    """The custom `Exception` class for Auth problems."""

    def __init__(
        self,
        status_code: int = 500,
        title: str = 'Unknown Error',
        detail: str = 'An unknown error occured.'
    ) -> None:

        self.__code = status_code
        self.__title = title
        self.__detail = detail

    @property
    def status_code(self) -> int:
        return self.__code

    @property
    def errors(self) -> list[dict[str, str]]:
        return [
            {
                'title': self.__title,
                'detail': self.__detail
            }
        ]


class ForbiddenError(AuthError):
    def __init__(self) -> None:
        super().__init__(
            status_code=403,
            title='Forbidden',
            detail="The user's role is not permitted to complete this action."
        )


class StateNotFoundError(AuthError):
    def __init__(self) -> None:
        super().__init__(
            400,
            'Unknown State',
            'The given state UUID is not recognised.'
        )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)
from ...utils import convert_s3_to_https


class TolqcDataToElasticRunDataConverter(DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        target_attributes = {}
        target_to_one = {}

        target_attributes['tag_index'] = data_object.tag_index
        target_attributes['tag_sequence'] = data_object.tag1_id
        target_attributes['tag2_sequence'] = data_object.tag2_id
        target_attributes['manual_qc'] = data_object.lims_qc
        target_attributes['auto_qc'] = data_object.auto_qc
        target_attributes['qc'] = data_object.qc
        target_attributes['read_length_n50'] = data_object.read_length_n50
        target_attributes['reads'] = data_object.reads
        target_attributes['bases'] = data_object.bases
        target_attributes['bases_a'] = data_object.bases_a
        target_attributes['bases_c'] = data_object.bases_c
        target_attributes['bases_g'] = data_object.bases_g
        target_attributes['bases_t'] = data_object.bases_t
        if data_object.library is not None \
                and data_object.library.library_type is not None:
            target_attributes['reporting_category'] = \
                data_object.library.library_type.reporting_category

        if data_object.run is not None:
            target_attributes['run'] = data_object.run.id
            target_attributes['position'] = data_object.run.element
            target_attributes['run_start'] = data_object.run.start
            target_attributes['run_complete'] = data_object.run.complete

            if data_object.run.platform is not None:
                target_attributes['instrument_model'] = data_object.run.platform.model

        if data_object.sample is not None:
            target_to_one['sequencing_request'] = self._data_object_factory(
                'sequencing_request',
                data_object.sample.id
            )
            if data_object.sample.specimen is not None:
                target_to_one['tolid'] = self._data_object_factory(
                    'tolid',
                    data_object.sample.specimen.id
                )
                if data_object.sample.specimen.supplied_name is not None:
                    target_to_one['specimen'] = self._data_object_factory(
                        'specimen',
                        data_object.sample.specimen.supplied_name
                    )
                if data_object.sample.specimen.accession is not None:
                    target_attributes['biospecimen_id'] = data_object.sample.specimen.accession.id
                if data_object.sample.specimen.species is not None:
                    target_to_one['species'] = self._data_object_factory(
                        'species',
                        data_object.sample.specimen.species.taxon_id
                    )

        if data_object.folder is not None and data_object.folder.image_file_list is not None:
            target_attributes['images'] = [
                {
                    'url':
                        convert_s3_to_https(
                            f'{data_object.folder.folder_location.uri_prefix}'
                            f'/{data_object.folder.id}'
                            '/' + file.get('file', '')
                        ),
                    'caption': file.get('caption')
                }
                for file in data_object.folder.image_file_list
            ]

        ret = self._data_object_factory(
            'run_data',
            data_object.id,
            attributes=target_attributes,
            to_one=target_to_one
        )
        yield ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import abstractmethod
from functools import cache

from .enum import (
    RelationWriteMode,
    ReturnMode
)


class _Writer:
    """
    A (private-API) class that documents:

    - how to perform writes for related
      instances of `DataObject`.
    - what types write methods return
    """

    @property
    @cache
    def write_mode(
        self
    ) -> dict[str, RelationWriteMode]:
        """
        The `WriteMode` value for this
        `object_type` on this `DataSource`
        instance.

        Override for custom behaviour.
        """

        return {
            type_: self._default_write_mode
            for type_ in self.supported_types
        }

    @property
    @cache
    def return_mode(
        self
    ) -> dict[str, ReturnMode]:
        """
        The `ReturnMode` value for this
        `object_type` on this `DataSource`
        instance.

        Override for custom behaviour.
        """

        return {
            type_: self._default_return_mode
            for type_ in self.supported_types
        }

    @property
    def _default_write_mode(
        self
    ) -> RelationWriteMode:
        """
        Unless `write_mode` is overriden, this
        is the value for all `supported_types`.
        """

        return RelationWriteMode.NOT_APPLICABLE

    @property
    def _default_return_mode(
        self
    ) -> ReturnMode:
        """
        Unless `return_mode` is overriden, this
        is the value for all `supported_types`.
        """

        return ReturnMode.NONE

    @property
    @abstractmethod
    def supported_types(self) -> list[str]:
        pass



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import datetime
from collections.abc import Callable, Iterable, Mapping
from dataclasses import dataclass
from typing import Any, TYPE_CHECKING

import dateutil

from .filter import ElasticFilterConverter
from ..core import DataObject, DataSourceFilter, DataSourceParser

if TYPE_CHECKING:
    from . import ElasticDataSource


ElasticApiResource = dict[str, Any]


@dataclass(slots=True)
class ElasticUpsertInputResource:
    index: str
    objects: Iterable[DataObject]
    id_func: Callable
    field_prefix: str


@dataclass(slots=True)
class ElasticUpdateInputResource:
    object_type: str
    update: dict
    field_prefix: str
    candidate_key: Iterable[str]


class DefaultElasticApiParser(DataSourceParser[ElasticApiResource, DataObject]):
    """
    Parses Elastic API transfer resource `dict`s to `DataObject` instances
    """
    __slots__ = ['__data_source']
    __data_source: ElasticDataSource

    def __init__(self, data_source: ElasticDataSource) -> None:
        self.__data_source = data_source

    def parse(self, transfer: ElasticApiResource) -> DataObject | None:
        if '_source' in transfer:
            type_ = self.__data_source._real_index_to_object_type(transfer['_index'])
            id_ = transfer['_id']
            attributes = transfer['_source']
            runtime_attributes = transfer['fields'] if 'fields' in transfer else {}
            return self._convert_data_dict_to_data_object(
                type_,
                id_,
                attributes,
                runtime_attributes
            )
        else:
            return None

    def _convert_data_dict_to_data_object(self, type_, id_, data, runtime_data):
        attributes = {
            k: self.__make_dates(type_, k, v) for k, v in data.items()
            if k in self.__data_source.attribute_types[type_]
        }
        runtime_attributes = {
            k: self.__make_dates(type_, k, v[0]) for k, v in runtime_data.items()
            if k in self.__data_source.attribute_types[type_]
        }
        to_one = self.__make_to_one_relations(type_, data)
        return self.__data_source.data_object_factory(
            type_,
            id_=id_,
            attributes=attributes | runtime_attributes,
            to_one=to_one
        )

    def __make_dates(self, object_type, attribute_name, value):
        if self.__data_source.attribute_types[object_type][attribute_name] == 'datetime' and \
                isinstance(value, str):
            return dateutil.parser.parse(value)
        return value

    def __make_to_one_relations(
        self,
        type_: str,
        data: dict[str, Any]
    ) -> dict[str, DataObject | None]:

        if type_ not in self.__data_source.relationship_config:
            return {}

        if self.__data_source.relationship_config[type_].to_one is None:
            return {}

        return {
            k: self.__make_to_one_relation(data.get(k), v)
            for k, v in self.__data_source.relationship_config[type_].to_one.items()
        }

    def __make_to_one_relation(
        self,
        relation_data: dict[str, Any] | None,
        type_: str
    ) -> DataObject | None:

        if (
            relation_data is None
            or not isinstance(relation_data, Mapping)
        ):
            return None

        id_ = relation_data.get('id')

        if id_ is None:
            return None

        return self._convert_data_dict_to_data_object(
            type_,
            id_,
            relation_data,
            {}  # This can be empty because runtime_fields are not applicable for enriched objects
        )


class _ToElasticApiResourceParser:
    def _convert_dates(self, dict_: dict) -> dict:
        ret = {}
        for k, v in dict_.items():
            if isinstance(v, datetime.datetime):
                ret[k] = v.isoformat()
            else:
                ret[k] = v
        return ret

    @property
    def _update_script(self):
        s = """
            for (param in params['upsertWith'].entrySet()) {
                if (param.value != null) {
                    if (ctx._source[param.key] instanceof Map) {
                        for (newParam in param.value.entrySet()) {
                            ctx._source[param.key][newParam.key] = newParam.value;
                        }
                        continue
                    }
                    if (ctx._source[param.key] instanceof ArrayList) {
                        for (newParam in param.value) {
                            if(! ctx._source[param.key].contains(newParam)) {
                                ctx._source[param.key].add(newParam)
                            }
                        }
                        continue
                    }
                }
                ctx._source[param.key] = param.value;
            }
        """
        return s.replace('\n', ' ')

    @property
    def _upsert_script(self):
        s = f"""
            if ( ctx.op == 'create' ) {{
                ctx._source = params['upsertWith']
            }} else {{
                {self._update_script}
            }}
        """
        return s.replace('\n', ' ')

    def _prefix_fields(self, dict_: dict, prefix: str) -> dict:
        if prefix == '':
            return dict_
        ret = {}
        for k, v in dict_.items():
            ret[prefix + '_' + k] = v
        return ret

    def _parse_to_one_relation(
        self,
        one_relation: DataObject | None
    ) -> dict[str, Any] | None:

        if one_relation is None:
            return None

        return {
            'id': one_relation.id,
            **one_relation.attributes
        }


class DefaultElasticUpsertInputParser(
    DataSourceParser[ElasticUpsertInputResource, ElasticApiResource],
    _ToElasticApiResourceParser
):
    __slots__ = ['__data_source']
    __data_source: ElasticDataSource

    def __init__(self, data_source: ElasticDataSource) -> None:
        self.__data_source = data_source

    def parse(
        self,
        transfer: ElasticUpsertInputResource,
    ):
        real_index_name = self.__data_source._get_indices().get(transfer.index)
        for object_ in transfer.objects:
            obj = self._convert_data_object_to_dict(object_)
            obj = self._convert_dates(obj)
            obj = self._prefix_fields(obj, transfer.field_prefix)
            obj = self._stringify_ids(obj)
            uid = transfer.id_func(object_)
            obj = self._add_uid(obj, uid)
            yield {
                '_op_type': 'update',
                'scripted_upsert': True,
                'upsert': {},
                '_index': real_index_name,
                '_id': uid,
                'script': {
                    'source': self._upsert_script,
                    'lang': 'painless',
                    'params': {
                        'upsertWith': obj
                    }
                }
            }

    def _convert_data_object_to_dict(self, data_object: DataObject) -> dict:
        to_ones_dict = {
            k: self._parse_to_one_relation(v)
            for k, v in data_object._to_one_objects.items()
        }
        return data_object.attributes | to_ones_dict

    def _stringify_ids(self, dict_: dict) -> dict:
        ret = {}
        for k, v in dict_.items():
            if isinstance(v, dict):
                if 'id' in v:
                    v['id'] = str(v['id'])
                ret[k] = self._stringify_ids(v)
            else:
                ret[k] = v

        return ret

    def _add_uid(self, dict_: dict, uid: Any) -> dict:
        return {**dict_, 'uid': f'{uid}'}


class DefaultElasticUpdateInputParser(
    DataSourceParser[ElasticUpdateInputResource, ElasticApiResource],
    _ToElasticApiResourceParser
):
    __slots__ = ['__data_source']
    __data_source: ElasticDataSource

    def __init__(self, data_source: ElasticDataSource) -> None:
        self.__data_source = data_source

    def parse(
        self,
        transfer: ElasticUpdateInputResource,
    ):
        u = self._convert_dates(transfer.update)
        f = DataSourceFilter()
        f.and_ = {}
        for key in transfer.candidate_key:
            # Don't want key in the upsert as it cannot change anyway
            f.and_[key] = {'eq': {'value': u.pop(key)}}
        u = self._prefix_fields(u, transfer.field_prefix)
        u = self._convert_data_objects_in_update_to_dict(u)
        query = ElasticFilterConverter(self.__data_source).convert(
            transfer.object_type, object_filters=f
        )
        return {
            'query': query,
            'script': {
                'source': self._update_script,
                'lang': 'painless',
                'params': {
                    'upsertWith': u
                }
            },
        }

    def _convert_data_objects_in_update_to_dict(self, dict_: dict) -> dict:
        ret = {}
        for k, v in dict_.items():
            if isinstance(v, DataObject):
                ret[k] = self._parse_to_one_relation(v)
            else:
                ret[k] = v
        return ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from datetime import datetime, timezone
from itertools import chain, groupby
from typing import Dict, Iterable, List, Optional, Type

from more_itertools import peekable

from tol.core import is_iter

from .core_converter import Converter
from .data_object import DataObject
from .data_object_converter import DataObjectToDataObjectOrUpdateConverter
from .datasource import DataSource
from .datasource_filter import DataSourceFilter


class DataLoader(ABC):
    @abstractmethod
    def load(self, provenance: str | None = None, dry_run: bool = False, **kwargs):
        """
        Loads a set of object from one DataSource to another
        """


class DefaultDataLoader():
    def __init__(
        self,
        source: DataSource,
        destination: DataSource,
        dependencies: List[Type['DataLoader']],
        source_object_type: str,
        loader_name: str,
        destination_object_type: str = '',
        audit: Optional[DataSource] = None,
        convert_class: Optional[DataObjectToDataObjectOrUpdateConverter] = None,
        object_filters: Optional[DataSourceFilter] = None,
        converter: Converter | None = None
    ):
        self._source = source
        self._destination = destination
        self._audit = audit
        self._dependencies = dependencies
        self._converter = self.__get_converter(convert_class, converter)
        self._converter.data_loader = self  # Set here to avoid circular import
        self._source_object_type = source_object_type
        self._destination_object_type = destination_object_type
        self._loader_name = loader_name
        self._object_filters = object_filters

    def load(
        self,
        provenance: str | None = None,
        dry_run: bool = False,
        candidate_key: Optional[List[str]] = ['id'],
        method: str = 'upsert',
        auto_exhaust: bool = True
    ):
        if not dry_run:
            self._record_time('start')

        source_objs = self._get_source_objects()
        converted_objs = self._convert_objects(source_objs, self._converter)

        if '' == self._destination_object_type:
            returned_objects = self._process_data_loads(
                candidate_key=candidate_key,
                converted_objs=converted_objs,
                dry_run=dry_run,
                provenance=provenance,
                method=method
            )
        else:
            returned_objects = self._process_data_load(
                candidate_key=candidate_key,
                converted_objs=converted_objs,
                dry_run=dry_run,
                provenance=provenance,
                method=method,
                destination_object_type=self._destination_object_type
            )

        if not dry_run:
            self._record_time('end')
            converter_return_object = self._check_converter_for_returned_objects()
            if is_iter(converter_return_object) and is_iter(returned_objects):
                returned_objects = chain(
                    returned_objects,
                    converter_return_object
                )

            if is_iter(returned_objects) and auto_exhaust:
                # Exhaust the returned objects
                for _ in returned_objects:
                    pass

            return returned_objects

    def _process_data_loads(
        self,
        converted_objs: Iterable[object],
        candidate_key: List[str] = ['id'],
        dry_run: bool = False,
        provenance: str | None = None,
        method: str = 'upsert'
    ):
        grouped_sorted_converted_objects = self._get_sorted_converted_objs(converted_objs)
        all_returned_objects = iter([])

        for key in grouped_sorted_converted_objects.keys():
            returned_objects = self._process_data_load(
                candidate_key=candidate_key,
                converted_objs=grouped_sorted_converted_objects[key],
                dry_run=dry_run,
                provenance=provenance,
                method=method,
                destination_object_type=key
            )

            all_returned_objects = chain(all_returned_objects, returned_objects)

        return all_returned_objects

    def _get_sorted_converted_objs(
            self,
            converted_objs: Iterable[object]
    ) -> Dict[str, Iterable[object]]:
        sorted_objects = sorted(converted_objs, key=lambda obj: obj.type)

        return {
            key: list(group) for key, group in groupby(sorted_objects, key=lambda obj: obj.type)
        }

    def _process_data_load(
        self,
        destination_object_type: str,
        converted_objs: Iterable[object],
        candidate_key: List[str] = ['id'],
        dry_run: bool = False,
        provenance: str | None = None,
        method: str = 'upsert'
    ):
        returned_objects = []
        if candidate_key == ['id']:
            if not dry_run:
                insert_method = getattr(self._destination, method)
                returned_objects = insert_method(
                    destination_object_type,
                    objects=converted_objs,
                    provenance=provenance
                )
            else:
                for converted_obj in converted_objs:
                    print(f'{converted_obj.id}: {converted_obj.attributes}')
        else:
            if not dry_run:
                returned_objects = self._destination.update(
                    object_type=destination_object_type,
                    updates=converted_objs,
                    candidate_key=candidate_key,
                    provenance=provenance
                )
            else:
                for converted_obj_id, converted_obj in converted_objs:
                    print(converted_obj_id, converted_obj)

        return returned_objects

    def __get_converter(
        self,
        convert_class: type[Converter] | None,
        converter: Converter | None
    ) -> Converter:

        if converter is not None:
            return converter

        if convert_class is not None:
            return convert_class(
                data_object_factory=self._destination.data_object_factory
            )

        raise Exception(
            'Either `convert_class` or `converter` must be specified.'
        )

    def _get_source_objects(self) -> Iterable:
        source_objs = self._source.get_list(
            self._source_object_type,
            object_filters=self._object_filters)
        return source_objs

    def _record_time(self, start_or_end: str):
        if self._audit is None:
            return
        new_datetime = datetime.now(timezone.utc)
        CoreDataObject = self._audit.data_object_factory  # noqa N806

        destination_object_type = 'multiple' \
            if self._destination_object_type == '' else self._destination_object_type

        audit_obj = CoreDataObject(
            'data_load_event',
            id_=self._loader_name,
            attributes={
                f'{start_or_end}_time': new_datetime,
                'source_object_type': self._source_object_type,
                'destination_object_type': destination_object_type
            }
        )
        self._audit.upsert('data_load_event', [audit_obj])

    def _convert_objects(self, objs: Iterable, _converter):
        converted_objects = _converter.convert_iterable(objs)

        return converted_objects

    def _check_converter_for_returned_objects(self):
        return_objects = self._converter.get_return_objects()

        if not 0 == len(return_objects):
            for return_object in return_objects:
                return_object.attributes['retrieved'] = True

            return iter(return_objects)

        return None


class GroupStatterDataLoader(DefaultDataLoader):

    def get_default_converter(self):
        # This will convert:
        # [{'key': {'field1': 'ID123', 'subfield': 'CAT'}
        #   'stats': {'count': 123, 'other': {'stat': 345}}},
        #  {'key': {'field1': 'ID123', 'subfield': 'DOG'}
        #   'stats': {'count': 456, 'other': {'stat': 678}}}
        # ]
        # to a CoreDataObject of type destination_object_type
        # with id: ID123
        # and attributes {'count_cat': 123,
        #                 'count_dog': 456,
        #                 'other_stat_cat': 345,
        #                 'other_stat_dog': 678}}
        data_loader = self

        class DefaultGroupStatToDataObjectConverter(DataObjectToDataObjectOrUpdateConverter):
            def convert_iterable(
                self,
                inputs: Iterable[DataObject]
            ) -> Iterable[DataObject]:
                peekable_inputs = peekable(inputs)
                additional_attributes = {}
                for input_ in peekable_inputs:
                    data_object = next(self.convert(input_))  # There will be one
                    # We might have some attributes here from before
                    for attr_name, attr_value in additional_attributes.items():
                        setattr(data_object, attr_name, attr_value)
                    next_stat = peekable_inputs.peek(None)
                    if next_stat is not None and \
                            next_stat['key'][data_loader._group_statter_group_by[0]] \
                            == data_object.id:
                        # We have more stats coming for this ID
                        additional_attributes = data_object.attributes
                    else:
                        additional_attributes = {}
                        yield data_object

            def convert(self, data_object: DataObject) -> Iterable[DataObject]:
                # It's not a DataObject, sorry!
                CoreDataObject = self._data_object_factory  # noqa N806
                source_object_type = data_loader._source_object_type
                # This gets the string to append
                append_string = '_'.join(data_object['key'][k]
                                         for k in data_loader._group_statter_group_by[1::])
                append_string = append_string.lower()
                if append_string != '':
                    append_string = f'{append_string}_'
                attributes = {}
                attributes[f'{source_object_type}_{append_string}count'] \
                    = data_object['stats']['count']
                for stats_field in data_loader._group_statter_stats_fields:
                    for stat in data_loader._group_statter_stats:
                        converted_stats_field = stats_field.replace('.', '_')
                        attributes[
                            f'{source_object_type}_{converted_stats_field}_{append_string}{stat}'
                        ] = data_object['stats'][stats_field][stat]
                ret1 = CoreDataObject(
                    id_=data_object['key'][data_loader._group_statter_group_by[0]],
                    type_=data_loader._destination_object_type,
                    attributes=attributes
                )
                yield ret1
        return DefaultGroupStatToDataObjectConverter

    def __init__(
        self,
        source: DataSource,
        destination: DataSource,
        dependencies: List[Type['DataLoader']],
        source_object_type: str, destination_object_type: str,
        loader_name: str,
        audit: DataSource = None,
        convert_class: Optional[DataObjectToDataObjectOrUpdateConverter] = None,
        object_filters: Optional[DataSourceFilter] = None,
        group_statter_group_by: Optional[str] = None,
        group_statter_stats_fields: Optional[List[str]] = [],
        group_statter_stats: Optional[List[str]] = ['min', 'max']
    ):
        if convert_class is None:
            convert_class = self.get_default_converter()
        super().__init__(
            source=source, destination=destination,
            dependencies=dependencies, source_object_type=source_object_type,
            destination_object_type=destination_object_type,
            loader_name=loader_name, audit=audit,
            convert_class=convert_class,
            object_filters=object_filters)
        self._group_statter_group_by = group_statter_group_by
        self._group_statter_stats_fields = group_statter_stats_fields
        self._group_statter_stats = group_statter_stats

    def _get_source_objects(self) -> Iterable:
        source_objs = self._source.get_group_stats(
            self._source_object_type,
            group_by=self._group_statter_group_by,
            stats_fields=self._group_statter_stats_fields,
            stats=self._group_statter_stats,
            object_filters=self._object_filters,
        )
        return source_objs


class IdsDataLoader(DefaultDataLoader):
    def __init__(self, source: DataSource, destination: DataSource,
                 dependencies: List[Type['DataLoader']],
                 source_object_type: str, destination_object_type: str,
                 loader_name: str,
                 audit: Optional[DataSource] = None,
                 convert_class: Optional[DataObjectToDataObjectOrUpdateConverter] = None,
                 object_ids: Optional[Iterable[str]] = None,
                 converter: Converter | None = None):
        super().__init__(
            source=source, destination=destination,
            dependencies=dependencies, source_object_type=source_object_type,
            destination_object_type=destination_object_type,
            loader_name=loader_name, audit=audit,
            converter=converter,
            convert_class=convert_class)
        self._object_ids = object_ids

    def _get_source_objects(self) -> Iterable:
        source_objs = self._source.get_by_ids(
            self._source_object_type,
            self._object_ids)
        return source_objs


class ObjectsDataLoader(DefaultDataLoader):
    def __init__(self, source: DataSource, destination: DataSource,
                 dependencies: List[Type['DataLoader']],
                 source_object_type: str, destination_object_type: str,
                 loader_name: str,
                 audit: Optional[DataSource] = None,
                 convert_class: Optional[DataObjectToDataObjectOrUpdateConverter] = None,
                 objects: Optional[Iterable[DataObject]] = None):
        super().__init__(
            source=source, destination=destination,
            dependencies=dependencies, source_object_type=source_object_type,
            destination_object_type=destination_object_type,
            loader_name=loader_name, audit=audit,
            convert_class=convert_class)
        self._objects = objects

    def _get_source_objects(self) -> Iterable:
        return self._objects



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class BenchlingSequencingRequestToElasticSequencingRequestConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        if data_object.sts_id and data_object.id:
            extraction = None
            extraction_container = None
            tissue_prep = None
            if 'extraction_id' in data_object.attributes:
                extraction = self._data_object_factory(
                    'extraction',
                    data_object.extraction_id
                )
            if 'fluidx_container_id' in data_object.attributes:
                extraction_container = self._data_object_factory(
                    'extraction_container',
                    data_object.fluidx_container_id
                )
            if 'tissue_prep_id' in data_object.attributes:
                tissue_prep = self._data_object_factory(
                    'tissue_prep',
                    data_object.tissue_prep_id
                )
            ret = self._data_object_factory(
                'sequencing_request',
                data_object.id,
                attributes={
                    **{k: v
                       for k, v in data_object.attributes.items()
                       if k not in ['sanger_sample_id', 'sts_id',
                                    'specimen_id', 'taxon_id', 'extraction_id',
                                    'programme_id', 'tissue_prep_id', 'fluidx_container_id']}
                },
                to_one={
                    'sample': self._data_object_factory(
                        'sample',
                        data_object.sts_id
                    ) if data_object.sts_id is not None else None,
                    'specimen': self._data_object_factory(
                        'specimen',
                        data_object.specimen_id
                    ) if data_object.specimen_id is not None else None,
                    'species': self._data_object_factory(
                        'species',
                        data_object.taxon_id
                    ) if data_object.taxon_id is not None else None,
                    'tolid': self._data_object_factory(
                        'tolid',
                        data_object.programme_id
                    ) if data_object.programme_id is not None else None,
                    'extraction': extraction,
                    'extraction_container': extraction_container,
                    'tissue_prep': tissue_prep,
                })
            yield ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from ..core import (
    core_data_object
)
from ..dummy import (
    DummyDataSource,
    create_dummy_datasource
)


def dummy(**kwargs) -> DummyDataSource:
    dummy = create_dummy_datasource()
    core_data_object(dummy)
    return dummy



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Dict, Optional, Tuple
from urllib.parse import quote

import requests

from .converter import GoatApiTransfer


class GoatApiClient:
    """
    Takes GoaT API transfers and connects to a remote
    GoaT API.
    """

    def __init__(
        self,
        goat_url: str,
    ) -> None:
        self.__goat_url = goat_url

    def get_detail(
        self,
        object_type: str,
        object_ids: str
    ) -> Optional[GoatApiTransfer]:
        """
        Gets a list of GoaT API transfers for the objects of specified
        `object_type` and `object_id`, or returns None if not found.
        """

        url, params = self.__detail_url(object_type, object_ids)
        return self.__fetch_detail(url, params)

    def get_list_page(
        self,
        object_type: str,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        filter_string: Optional[str] = None,
        sort_by: Optional[str] = None
    ) -> Tuple[GoatApiTransfer, int]:
        url, params = self.__page_url(object_type, page, page_size, filter_string, sort_by)
        return self.__fetch_detail(url, params)

    def __fetch_detail(
        self,
        url: str,
        params: Dict = {}
    ) -> Optional[GoatApiTransfer]:
        r = requests.get(url, params=params)
        if r.status_code == 404:
            return [], 0
        r.raise_for_status()
        return r.json()['results'] if 'results' in r.json() else [], r.json()['status']['hits']

    def __detail_url(self, object_type: str, object_ids: list[str]) -> Tuple[str, Dict]:
        obj_ids_str = ','.join(object_ids)
        return self.__page_url(object_type, 1, len(object_ids), f'tax_name({obj_ids_str})')

    def __page_url(
        self,
        object_type: str,
        page: int,
        page_size: int,
        filter_string: Optional[str] = None,
        sort_by: Optional[str] = None
    ) -> Tuple[str, Dict]:
        url = f'{self.__goat_url}/search?query={quote(filter_string)}'
        params = {
            # 'query': filter_string,  # We encode the query string manually to ensure space->%20
            'result': object_type,
            'includeEstimates': 'true',
            'summaryValues': 'count',
            'taxonomy': 'ncbi',
            'names': 'common_name,tolid_prefix,synonym',
            'fields': 'genome_size,chromosome_number,assembly_level,haploid_number,ploidy,'
                      'echabs92,habreg_2017,marhabreg-2017,waca_1981,isb_wildlife_act_1976,'
                      'protection_of_badgers_act_1992,family_representative,long_list,'
                      'sample_collected,country_list',
            'ranks': 'subspecies,species,genus,family,superfamily,order,class,phylum,kingdom,'
                     'superkingdom,domain'
        } | self.__get_sort_params(sort_by) | self.__get_page_params(page, page_size)
        return url, params

    def __get_page_params(self, page: int, page_size: int) -> Dict:
        ret = {}
        if page is not None:
            ret['offset'] = (page - 1) * page_size
        if page_size is not None:
            ret['size'] = page_size
        return ret

    def __get_sort_params(self, sort_by: str) -> Dict:
        if sort_by is None:
            return {}
        elif sort_by.startswith('-'):
            return {
                'sortBy': sort_by[1:] if sort_by[1:] != 'id' else 'taxon_id',
                'sortOrder': 'desc'
            }
        else:
            return {
                'sortBy': sort_by if sort_by != 'id' else 'taxon_id',
                'sortOrder': 'asc'
            }



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
# SPDX-License-Identifier: MIT

import re
from dataclasses import dataclass
from typing import Iterable

from tol.core import DataObject, DataObjectToDataObjectOrUpdateConverter


class AutoDetectManifestTypeConverter(DataObjectToDataObjectOrUpdateConverter):

    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        rack_or_plate: str

    __slots__ = ['__config']
    __config: Config

    def __init__(self, data_object_factory, config: Config) -> None:
        super().__init__(data_object_factory)
        self.__config = config
        self._data_object_factory = data_object_factory

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        """
        converting the samples DataObject into ENA format
        """
        s = data_object
        attributes = {}
        rack_or_plate_based_manifest = bool(
            re.fullmatch(r'^[A-P][12]?[0-9]$',
                         s.attributes.get(self.__config.rack_or_plate))
        )

        if rack_or_plate_based_manifest:
            attributes['manifest_type'] = 'PLATE_WELL'
        else:
            attributes['manifest_type'] = 'RACK_TUBE'

        ret = self._data_object_factory(
            data_object.type,
            s.id,
            attributes=attributes,
        )
        yield ret



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Iterable

from tol.core import DataObject, DataObjectToDataObjectOrUpdateConverter


class SkipNullFieldsConverter(DataObjectToDataObjectOrUpdateConverter):

    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        field_names: list[str]

    __slots__ = ['__config']
    __config: Config

    def __init__(self, data_object_factory, config: Config) -> None:
        super().__init__(data_object_factory)
        self.__config = config
        self._data_object_factory = data_object_factory

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        """
        removing null fields from the DataObject
        """

        passes = True
        for field in self.__config.field_names:
            value = data_object.get_field_by_name(field)
            if value is None:
                passes = False
                break

        if passes:
            ret = self._data_object_factory(
                data_object.type,
                data_object.id,
                attributes=data_object.attributes
            )
            yield ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
import os

from tol.core import (
    core_data_object
)
from tol.google_sheets import (
    GoogleSheetDataSource
)


def portal_attributes(**kwargs) -> GoogleSheetDataSource:
    gsds = GoogleSheetDataSource({
        'client_secrets': json.loads(os.getenv('GOOGLE_CLIENT_SECRETS')),
        'sheet_key': '1kKta1ziEqAqXd5-ZJmv3PA9lE4kxh6BD4Bpxh0HhxeQ',
        'mappings': {
            'attribute': {
                'worksheet_name': 'Attributes',
                'columns': {
                    'object_type': {
                        'heading': 'object_type',
                        'type': 'str'
                    },
                    'id': {
                        'heading': 'id',
                        'type': 'str'
                    },
                    'name': {
                        'heading': 'name',
                        'type': 'str'
                    },
                    'display_name': {
                        'heading': 'display_name',
                        'type': 'str'
                    },
                    'authoritative': {
                        'heading': 'authoritative',
                        'type': 'boolean'
                    },
                    'available_on_relationships': {
                        'heading': 'available_on_relationships',
                        'type': 'boolean'
                    },
                    'description': {
                        'heading': 'description',
                        'type': 'str'
                    },
                    'source': {
                        'heading': 'source',
                        'type': 'str'
                    }
                },
                'header_row': 5,
                'data_start_row': 6
            }
        }
    })
    core_data_object(gsds)
    return gsds



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from datetime import datetime
from itertools import chain
from typing import Optional

from .converter import CopoApiTransfer
from ..core import HttpClient


class CopoApiClient(HttpClient):
    """
    Takes COPO API transfers and connects to a remote
    COPO API.
    """

    def __init__(
        self,
        copo_url: str,
    ) -> None:
        super().__init__()
        self.__copo_url = copo_url

    def get_detail(
        self,
        object_type: str,
        object_ids: str
    ) -> Optional[CopoApiTransfer]:
        """
        Gets a single COPO API transfer for the object of specified
        `object_type` and `object_id`, or returns None if not found.
        """

        urls = self.__detail_url(object_type, object_ids)
        return chain.from_iterable(
            self.__fetch_detail(url, object_type, id_)
            for id_, url in urls
        )

    def get_samples_in_manifest(
        self,
        manifest_id: str
    ) -> Optional[CopoApiTransfer]:
        """
        Gets a list of COPO API transfers for the samples in a
        specified manifest, or returns None if not found.
        """

        url = f'{self.__copo_url}/manifest/{manifest_id}'
        return self.__fetch_detail(
            url,
            'samples_in_manifest',  # A fudge to ignore the manifest-specific stuff
            manifest_id
        )

    def get_manifests(
        self,
        project: str,
        from_: datetime,
        to: datetime
    ) -> Optional[CopoApiTransfer]:
        """
        Gets a list of COPO API transfers for the samples in a
        specified manifest, or returns None if not found.
        """

        from_string = from_.isoformat()
        to_string = to.isoformat()
        url = f'{self.__copo_url}/manifest/{project}/{from_string}/{to_string}'
        manifest_ids = self.__fetch_detail(
            url,
            'manifests_between_dates'  # A fudge to ignore the manifest-specific stuff
        )
        return [
            {
                'tolsdk-type': 'manifest',
                'copo_id': id_
            }
            for id_ in manifest_ids
        ]

    def __fetch_detail(
        self,
        url: str,
        object_type: str,
        object_id: str = None
    ) -> Optional[CopoApiTransfer]:
        print(f'Getting {url}')
        session = self._get_session_with_retries()
        r = session.get(url, headers={'accept-encoding': 'gzip;q=0,deflate,sdch'})
        if r.status_code in [400, 404]:
            return [None]
        r.raise_for_status()
        if object_type == 'manifest':
            # The manifest does not exist if the response's data is
            # a list of strings
            if len(r.json()['data']) > 0 and isinstance(r.json()['data'][0], str):
                return [None]
            ret = [{'tolsdk-type': object_type, 'copo_id': object_id}]
        else:
            ret = r.json()['data']
        return ret

    def __detail_url(self, object_type: str, object_ids: str) -> str:
        # It is possible to request multiple sample IDs on the same request,
        # but if one of them is invalid, the request will return a 400 status code.
        # Therefore, we make a request for each sample ID.
        path = f'{object_type}'
        if object_type == 'sample':
            path = f'{object_type}/copo_id'
        return [
            (id_, f'{self.__copo_url}/{path}/{id_}')
            for id_ in object_ids
        ]



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Iterable, Optional

from ._filterable import _Filterable

if typing.TYPE_CHECKING:
    from ..data_object import DataObject
    from ..datasource_filter import DataSourceFilter
    from ..requested_fields import ReqFieldsTree
    from ..session import OperableSession


class ListGetter(_Filterable, ABC):
    """
    Gets an Iterable of DataObject instances.
    """

    @abstractmethod
    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None,
        requested_fields: list[str] | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[DataObject]:
        """
        Gets an Iterable of DataObject instances of the given
        type, according to the given DataSourceFilter.
        """



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from functools import cache
from typing import Iterable, Optional

from cachetools.func import ttl_cache

import requests

from ..core import (
    DataObject,
    DataSource,
    DataSourceFilter
)
from ..core.operator import (
    PageGetter
)

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession


class StatusDataSource(
    DataSource,

    # the supported operators
    PageGetter,
):
    """
    A `DataSource` that gets status from a remote URL
    """
    @ttl_cache(ttl=60)
    def get_status(self, url) -> dict:
        try:
            r = requests.head(url, verify=False)
            return r.ok, r.status_code
        except Exception:
            return False, 500

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            'status': {
                'ok': 'bool',
                'status_code': 'int'
            }
        }

    @property
    @cache
    def supported_types(self) -> list[str]:
        return list(self.attribute_types.keys())

    def get_list_page(
        self,
        object_type: str,
        page_number: int,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        sort_by: Optional[str] = None,
        session: Optional[OperableSession] = None
    ) -> tuple[Iterable[DataObject], int]:

        urls = object_filters.and_['id']['in_list']['value']

        ret = []
        for url in urls:
            ok, status_code = self.get_status(url)
            ret.append(self.data_object_factory(
                'status',
                url,
                attributes={
                    'ok': ok,
                    'status_code': status_code
                }
            ))

        return ret, len(ret)



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from collections.abc import Mapping
from itertools import chain
from typing import Dict, Iterable, Tuple

from .datasource_error import UnknownObjectTypeException

if typing.TYPE_CHECKING:
    from .datasource import DataSource


class DataSourceDict(Mapping):
    """
    A useful implementation of a Dict[str, DataSource], mapping
    a type of DataObject to its "providing" DataSource, to be injected
    into the constructor of each Controller instance.

    A singleton can be used.
    """

    def __init__(self, *data_sources: DataSource) -> None:
        self.__data_sources = data_sources

    def __getitem__(self, object_type: str) -> DataSource:
        data_source = self.__get_dict().get(object_type)
        if data_source is None:
            raise UnknownObjectTypeException(object_type)
        return data_source

    def __setitem__(self, *args, **kwargs) -> None:
        raise NotImplementedError('This Dict is readonly.')

    def __len__(self) -> int:
        return len(self.__get_dict())

    def __iter__(self) -> Iterable:
        return iter(self.__get_dict())

    def __get_dict(self) -> Dict[str, DataSource]:
        object_tuples = [
            self.__get_pairs_iterable(data_source)
            for data_source in self.__data_sources
        ]
        return dict(chain(*object_tuples))

    def __get_pairs_iterable(
        self,
        data_source: DataSource
    ) -> Iterable[Tuple[str, DataSource]]:
        return [
            (object_type, data_source)
            for object_type in data_source.supported_types
        ]



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod

from .mapper import JiraMapper
from ..core import DataSourceFilter


class JiraFilter(JiraMapper, ABC):
    """
    Converts instances of `DataSourceFilter` to a
    valid Jira JQL dump.
    """

    @abstractmethod
    def dumps(self, filter_: DataSourceFilter) -> str:
        """
        Emit a filter string from a `DataSourceFilter` instance
        """
        pass


class DefaultJiraFilter(JiraFilter):

    def __init__(
        self,
        field_mappings: dict[str, dict[str, str]]
    ) -> None:
        super().__init__(field_mappings)

    __op_mappings = {
        'eq': ['=', '!='],
        'lt': ['<', '>='],
        'lte': ['<=', '>'],
        'gt': ['>', '<='],
        'gte': ['>=', '<'],
        'contains': ['~', '!~'],
        'in_list': ['in', 'not in'],
        'exists': ['is not null', 'is null'],
    }

    def dumps(self, filter_: DataSourceFilter) -> str:
        filter_conditions = []
        for field, filter_list in filter_.and_.items():
            for op, ops in self.__op_mappings.items():
                if op in filter_list:
                    specific_filter = filter_list[op]
                    if 'value' in specific_filter:
                        operator = ops[0]
                        if 'negate' in specific_filter:
                            operator = ops[1]
                        if op == 'in_list':
                            filter_condition = ','.join(
                                f'"{v}"' for v in specific_filter['value']
                            )
                            filter_condition = f'({filter_condition})'
                        else:
                            filter_condition = '"' + str(specific_filter['value']) + '"' \
                                if 'value' in specific_filter else ''
                        filter_conditions.append(f'{self._map_field(field)} {operator} '
                                                 f'{filter_condition}')
        filter_string = ' AND '.join(filter_conditions)
        return filter_string if filter_string is not None else ''



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Dict

from tol.core import Validator
from tol.core.data_object import DataObject


class SpecimensHaveSameTaxonValidator(Validator):
    """
    Validates an incoming stream of `DataObject` instances.
    For each data object (sample) not a SYMBIONT, it checks that
    there are no samples with SPECIMEN_ID which has different TAXON_ID
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        taxon_id_field: str
        symbiont_field: str
        specimen_id_field: str

    __slots__ = ['__config', '__seen']
    __config: Config
    __seen: Dict[str, str]

    def __init__(self, config: Config, **kwargs) -> None:
        super().__init__()
        self.__seen = {}
        self.__config = config

    def _validate_data_object(self, obj: DataObject) -> None:
        # Explaining the code concept using a standard example
        # seen{}
        # 1st Pass=>    element['specimen_id']  =   A
        #               element['taxon_id']     =   AA
        #               seen{ A:AA }
        # 2nd pass=>    element['specimen_id']  =   A
        #               element['taxon_id']     =   AB
        #               AB != AA
        #               Flag error
        # From Nithin :)

        # Ensure the data object is not a SYMBIONT
        if obj.attributes.get(self.__config.symbiont_field) != 'SYMBIONT':
            specimen_id = obj.attributes.get(self.__config.specimen_id_field)
            if specimen_id is None:
                return
            taxon_id = obj.attributes.get(self.__config.taxon_id_field)
            if taxon_id is None:
                return
            if specimen_id in self.__seen and taxon_id != self.__seen[specimen_id]:
                self.add_error(
                    object_id=obj.id,
                    detail='All samples from the same specimen_ID must have the same TAXON_ID.',
                    field=self.__config.specimen_id_field,
                )
            if specimen_id not in self.__seen:
                self.__seen[specimen_id] = taxon_id



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class MlwhSequencingRequestToElasticSequencingRequestConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        attributes = {
            k: v
            for k, v in data_object.attributes.items()
            if k not in ['taxon_id', 'public_name', 'sample_ref', 'supplier_name']
        }
        to_one_relations = {
            'specimen': self._data_object_factory(
                'specimen',
                data_object.supplier_name),
            'species': self._data_object_factory(
                'species',
                str(data_object.taxon_id)
            ),
            'tolid': self._data_object_factory(
                'tolid',
                data_object.public_name
            )
        }
        ret = self._data_object_factory(
            'sequencing_request',
            data_object.sample_ref,
            attributes=attributes,
            to_one=to_one_relations
        )
        yield ret



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import re
from dataclasses import dataclass
from typing import Any, List

from tol.core import DataObject
from tol.core.validate import Validator


RegexDict = dict[
    str,
    str | bool | list[Any],
]


@dataclass(frozen=True, kw_only=True)
class Regex:
    key: str
    regex: str

    is_error: bool = True
    detail: str = 'Value is not allowed for given key'

    def is_allowed(self, __v: Any) -> bool:
        # Check regex
        return __v is None or __v == '' or bool(re.search(
            self.regex,
            str(__v)
        ))


class RegexValidator(Validator):
    """
    Validates an incoming stream of `DataObject` instances
    according to the specified allowed values for a given
    key.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        regexes: List[Regex | RegexDict]

    __slots__ = ['__config']
    __config: Config

    def __init__(
        self,
        config: Config,
        **kwargs
    ) -> None:

        super().__init__()

        self.__config = self.__get_config(config)

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:

        for k, v in obj.attributes.items():
            self.__validate_attribute(obj, k, v)

    def __validate_attribute(
        self,
        obj: DataObject,
        key: str,
        value: Any,
    ) -> None:

        config = self.__filter_config(key)

        for c in config:
            if not c.is_allowed(value):
                self.__add_result(obj, c)

    def __filter_config(
        self,
        key: str,
    ) -> list[Regex]:
        return [
            a for a in self.__config.regexes
            if a.key == key
        ]

    def __add_result(
        self,
        obj: DataObject,
        c: Regex,
    ) -> None:
        if c.is_error:
            self.add_error(
                object_id=obj.id,
                detail=c.detail,
                field=c.key
            )
        else:
            self.add_warning(
                object_id=obj.id,
                detail=c.detail,
                field=c.key,
            )

    def __get_config(
        self,
        config: Config,
    ) -> Config:

        # Ensure config is in Regex format
        # (as you can either pass in a list of Regex or a RegexDict,
        # which can be used to initialize a Regex)
        return self.Config(
            regexes=[
                c if isinstance(c, Regex) else Regex(**c)
                for c in config.regexes
            ]
        )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Any, List, Optional

from .parser import Parser
from ..core import DataObject


DummyObject = dict[str, Any]
DummyTransfer = List[DummyObject]


class DummyConverter():

    """
    Converts from Dummy transfers to instances of
    `DataObject`.
    """

    def __init__(
        self,
        parser: Parser
    ) -> None:

        self.__parser = parser

    def convert(self, input_: DummyTransfer) -> DataObject:
        """
        Converts a DummyTransfer containing a detail (single) result
        """
        return self.__parser.parse(input_)

    def convert_list(
        self,
        input_: DummyTransfer
    ) -> tuple[list[DataObject], Optional[int]]:
        """
        Converts a DummyTransfer containing a list of results. Also
        returns a count of the total results meeting.
        """

        return [
            self.__parser.parse(json_obj)
            for json_obj in input_
            if json_obj is not None
        ], None



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from datetime import datetime
from typing import Any, Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)
from ...core.data_object import ErrorObject


class BenchlingSampleCasmToStsSampleConverter(DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        id_, attributes_ = self.__get_sample_upsert_id_and_attributes(data_object)
        ret = self._data_object_factory(
            'sample',
            str(id_),
            attributes=attributes_
        )
        yield ret

    def __get_sample_upsert_id_and_attributes(
        self,
        obj: DataObject | ErrorObject
    ) -> [str, dict[str, Any]]:

        if isinstance(obj, ErrorObject):
            return obj.object_.id, {
                'eln_error': {
                    'details': obj.details,
                    'object_type': obj.object_type,
                }
            }
        else:
            meta_data_object = obj.sample_metadata_id
            return meta_data_object.sts_id, {
                'eln_id': obj.id,
                'eln_updated_at': datetime.now()
            }



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from typing import Any

from ..core import DataSourceFilter, DataSourceFilterConverter

# Import only for typing due to circular imports
if typing.TYPE_CHECKING:
    from .elastic_datasource import ElasticDataSource


class ElasticFilterConverter(DataSourceFilterConverter):
    __slots__ = ['__elastic_datasource']
    __elastic_datasource: ElasticDataSource

    def __init__(self, elastic_datasource: ElasticDataSource) -> None:
        super().__init__()
        self.__elastic_datasource = elastic_datasource

    def convert(self, object_type: str, object_filters: DataSourceFilter | None = None) -> dict:
        query = {'bool': {'must': [], 'must_not': []}}
        object_filters = self.__elastic_datasource._preprocess_filter(object_type, object_filters)
        if object_filters is None:
            return query
        if object_filters.and_ is not None:
            for k, v in object_filters.and_.items():
                search_field = self.__elastic_datasource._field_or_keyword(object_type, k)
                for op, constraint in v.items():
                    search_value = constraint.get('value')
                    negated = constraint.get('negate', False)
                    elastic_section = 'must_not' if negated else 'must'
                    if 'field' in constraint:
                        other_field = self.__elastic_datasource._field_or_keyword(
                            object_type, constraint['field']
                        )
                        query['bool']['filter'] = \
                            self._get_field_comparison_filter(
                                search_field,
                                other_field,
                                op,
                                negated
                        )
                        continue
                    if op in ['gt', 'gte', 'lt', 'lte']:
                        query['bool'][elastic_section].append({
                            'range': {search_field: {op: search_value}}
                        })
                    if op in ['eq']:
                        query = self._eq_filter(
                            query,
                            elastic_section,
                            search_field,
                            search_value
                        )
                    if op in ['contains']:
                        query['bool'][elastic_section].append({
                            'wildcard': {
                                search_field: {'value': f'{search_value}*', 'boost': 1.0}
                            }
                        })
                    if op in ['exists']:
                        query['bool'][elastic_section].append({
                            'exists': {'field': search_field}
                        })
                    if op in ['in_list']:
                        query['bool'][elastic_section].append({
                            'terms': {search_field: search_value, 'boost': 1.0}
                        })
        return query

    def _get_field_comparison_filter(self, field1: str, field2: str, op: str, negated: bool) -> \
            dict[str, dict[str, str]]:
        op_mappings = {
            'eq': '==',
            'lt': '<',
            'lte': '<=',
            'gt': '>',
            'gte': '>='
        }
        negated_mappings = {  # What to return if negated
            True: 'true',
            False: 'false'
        }
        # return {negated_mappings[not negated]}
        return {
            'script': {
                'script': {
                    'source': f"""
                        if (doc[params['field1']].size() > 0
                            && doc[params['field2']].size() > 0) {{
                            if (doc[params['field1']].value.compareTo(doc[params['field2']].value)
                                {op_mappings[op]} 0) {{
                                return {negated_mappings[not negated]}
                            }}
                        }}
                        return {negated_mappings[negated]};
                    """,
                    'params': {
                        'field1': field1,
                        'field2': field2
                    }
                }
            }
        }

    def _eq_filter(
        self,
        query: dict[str, Any],
        elastic_section: str,
        search_field: str,
        search_value: str
    ) -> dict[str, Any]:

        query['bool'][elastic_section].append({
            'match': {search_field: search_value}
        })

        return query


def _build_elasticsearch_query(
    self,
    object_type: str,
    object_filters: DataSourceFilter | None = None
) -> dict:
    query = {'bool': {'must': [], 'must_not': []}}
    object_filters = self._preprocess_filter(object_type, object_filters)
    # If we want to implement preprocessing of filters, call self._preprocess_filter() here
    if object_filters is None:
        return query
    if object_filters.and_ is not None:
        for k, v in object_filters.and_.items():
            search_field = self._field_or_keyword(object_type, k)
            for op, constraint in v.items():
                search_value = constraint.get('value')
                negated = constraint.get('negate', False)
                elastic_section = 'must_not' if negated else 'must'
                if 'field' in constraint:
                    other_field = self._field_or_keyword(
                        object_type, constraint['field']
                    )
                    query['bool']['filter'] = \
                        self._get_field_comparison_filter(
                            search_field,
                            other_field,
                            op,
                            negated
                    )
                    continue
                if op in ['gt', 'gte', 'lt', 'lte']:
                    query['bool'][elastic_section].append({
                        'range': {search_field: {op: search_value}}
                    })
                if op in ['eq']:
                    query = self._eq_filter(
                        query,
                        elastic_section,
                        search_field,
                        search_value
                    )
                if op in ['contains']:
                    query['bool'][elastic_section].append({
                        'wildcard': {
                            search_field: {'value': f'{search_value}*', 'boost': 1.0}
                        }
                    })
                if op in ['exists']:
                    query['bool'][elastic_section].append({
                        'exists': {'field': search_field}
                    })
                if op in ['in_list']:
                    query['bool'][elastic_section].append({
                        'terms': {search_field: search_value, 'boost': 1.0}
                    })
    return query



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class GoatTaxonToElasticSpeciesConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        if data_object is not None and data_object.id is not None:
            # Is the ID an integer?
            try:
                int(data_object.id)
            except ValueError:
                return iter([])
            ret = self._data_object_factory(
                'species',
                data_object.id,
                attributes={
                    **data_object.attributes
                }
            )
            for rank in data_object._host.get_ranks():
                rank_object = getattr(data_object, rank)
                if rank_object is not None:
                    setattr(ret, f'{rank}_name', rank_object.scientific_name)
            yield ret
        return iter([])



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from ..core import (
    core_data_object
)
from ..mlwh import (
    MlwhDataSource
)


def mlwh(**kwargs) -> MlwhDataSource:
    mlwh = MlwhDataSource({'uri': os.getenv('MLWH_URI')})
    core_data_object(mlwh)
    return mlwh



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .defaults import Defaults
from ..bold import (
    BoldDataSource,
    create_bold_datasource
)
from ..core import (
    core_data_object
)


def bold(**kwargs) -> BoldDataSource:
    bold = create_bold_datasource(
        bold_url=os.getenv('BOLD_URL', Defaults.BOLD_URL)
        + os.getenv('BOLD_API_PATH', Defaults.BOLD_API_PATH),
        bold_portal_url=os.getenv('BOLD_PORTAL_URL', Defaults.BOLD_PORTAL_URL)
        + os.getenv('BOLD_API_PATH', Defaults.BOLD_API_PATH),
        bold_api_key=os.getenv('BOLD_API_KEY')
    )
    core_data_object(bold)
    return bold
